# -*- coding: utf-8 -*-
"""
policy.py

Validation routines related to the policy file
"""
from olxcleaner.utils import traverse
from olxcleaner.parser.parser_exceptions import (
    MissingURLName,
    DuplicateURLName,
    BadPolicyFormat,
    PolicyRefNotFound,
    WrongObjectType,
    BadEntry,
    SettingOverride,
    GradingPolicyIssue
)

def find_url_names(course, errorstore):
    """
    Constructs a dictionary of (url_name: EdxObject) references

    :param course: EdxCourse object with a loaded course
    :param errorstore: ErrorStore object where errors are reported
    :return: Dictionary of {'url_name': EdxObject} links
    """
    results = {}

    # Traverse the tree
    for edxobj in traverse(course):
        url_name = edxobj.attributes.get('url_name')

        if url_name is None:
            if edxobj.needs_url_name and not edxobj.broken:
                # Report the error
                errorstore.add_error(MissingURLName(edxobj.filenames[0], edxobj=edxobj))
        else:
            # Record the name
            if url_name in results:
                # We have a collision!
                errorstore.add_error(DuplicateURLName(edxobj.filenames[0],
                                                      url_name=url_name,
                                                      tag1=results[url_name].type,
                                                      file1=results[url_name].filenames[0],
                                                      tag2=edxobj.type,
                                                      file2=edxobj.filenames[0]))
            else:
                results[url_name] = edxobj

    # Return the dictionary
    return results

def merge_policy(policy, url_names, errorstore):
    """
    Merges policy file data with course objects

    :param policy: Policy file json
    :param url_names: Dictionary of {urlname: EdxObject} pairings
    :param errorstore: ErrorStore object where errors are reported
    :return: None
    """
    if not isinstance(policy, dict):
        errorstore.add_error(BadPolicyFormat('policy.json'))
        return

    for entry in policy:
        # Split the entry into object type/url_name
        objtype, url_name = entry.split("/", 1)

        # Find the corresponding object in url_names
        edxobj = url_names.get(url_name, None)

        # Make sure we have an object
        if edxobj is None:
            errorstore.add_error(PolicyRefNotFound('policy.json', objtype=objtype, url_name=url_name))
            continue

        # Make sure the type matches
        if edxobj.type != objtype:
            errorstore.add_error(WrongObjectType('policy.json', objtype=objtype, objtypefound=edxobj.type, url_name=url_name))
            continue

        # Make sure the entry is a dictionary
        if not isinstance(policy[entry], dict):
            errorstore.add_error(BadEntry('policy.json', objtype=objtype, url_name=url_name))
            continue

        # Copy data into the object
        for element in policy[entry]:
            # Make sure we're not overwriting anything
            if element in edxobj.attributes:
                errorstore.add_error(SettingOverride('policy.json', objtype=objtype, url_name=url_name, setting=element))
                continue

            # Copy the data
            edxobj.attributes[element] = policy[entry][element]

def validate_grading_policy(grading_policy, errorstore):
    """
    Validates the grading policy of the course

    :param grading_policy: Grading policy file json
    :param errorstore: ErrorStore object where errors are reported
    :return: None
    """
    # Check the GRADER entry
    if "GRADER" not in grading_policy:
        errorstore.add_error(GradingPolicyIssue('grading_policy.json', msg="'GRADER' entry not found in grading policy"))
    elif not isinstance(grading_policy["GRADER"], list):
        errorstore.add_error(GradingPolicyIssue('grading_policy.json', msg="'GRADER' entry not in grading policy is not a list"))
    else:
        weightsum = 0
        types = set()
        # Validate the required components of each entry
        for idx, entry in enumerate(grading_policy["GRADER"]):
            if "drop_count" not in entry:
                msg = f"'drop_count' setting is omitted for entry {idx+1} in the grading policy"
                errorstore.add_error(GradingPolicyIssue('grading_policy.json', msg=msg))
            elif not isinstance(entry['drop_count'], int):
                msg = f"'drop_count' setting is not an integer for entry {idx+1} in the grading policy"
                errorstore.add_error(GradingPolicyIssue('grading_policy.json', msg=msg))
            elif entry['drop_count'] < 0:
                msg = f"'drop_count' setting is negative for entry {idx+1} in the grading policy"
                errorstore.add_error(GradingPolicyIssue('grading_policy.json', msg=msg))

            if "min_count" not in entry:
                msg = f"'min_count' setting is omitted for entry {idx+1} in the grading policy"
                errorstore.add_error(GradingPolicyIssue('grading_policy.json', msg=msg))
            elif not isinstance(entry['min_count'], int):
                msg = f"'min_count' setting is not an integer for entry {idx+1} in the grading policy"
                errorstore.add_error(GradingPolicyIssue('grading_policy.json', msg=msg))
            elif entry['min_count'] < 1:
                msg = f"'min_count' setting is less than 1 for entry {idx+1} in the grading policy"
                errorstore.add_error(GradingPolicyIssue('grading_policy.json', msg=msg))

            if "short_label" in entry and not isinstance(entry['short_label'], str):
                msg = f"'short_label' setting is not a string for entry {idx+1} in the grading policy"
                errorstore.add_error(GradingPolicyIssue('grading_policy.json', msg=msg))

            if "type" not in entry:
                msg = f"'type' setting is omitted for entry {idx+1} in the grading policy"
                errorstore.add_error(GradingPolicyIssue('grading_policy.json', msg=msg))
            elif not isinstance(entry['type'], str):
                msg = f"'type' setting is not a string for entry {idx+1} in the grading policy"
                errorstore.add_error(GradingPolicyIssue('grading_policy.json', msg=msg))
            else:
                if entry['type'] in types:
                    msg = f"Assessment type '{entry['type']}' appears multiple times in the grading policy"
                    errorstore.add_error(GradingPolicyIssue('grading_policy.json', msg=msg))
                else:
                    types.add(entry['type'])

            if "weight" not in entry:
                msg = f"'weight' setting is omitted for entry {idx+1} in the grading policy"
                errorstore.add_error(GradingPolicyIssue('grading_policy.json', msg=msg))
            elif not (isinstance(entry['weight'], int) or isinstance(entry['weight'], float)):
                msg = f"'weight' setting is not a number between 0 and 1 for entry {idx+1} in the grading policy"
                errorstore.add_error(GradingPolicyIssue('grading_policy.json', msg=msg))
            elif entry['weight'] > 1 or entry['weight'] < 0:
                msg = f"'weight' setting is not a number between 0 and 1 for entry {idx+1} in the grading policy"
                errorstore.add_error(GradingPolicyIssue('grading_policy.json', msg=msg))
            else:
                weightsum += entry['weight']

        # Check that weights add up to 100%
        if len(grading_policy["GRADER"]) == 0:
            msg = "GRADER entry in grading policy should not be empty"
            errorstore.add_error(GradingPolicyIssue('grading_policy.json', msg=msg))
        elif abs(weightsum - 1) > 1e-8:
            msg = "'weight' settings do not add up to 1"
            errorstore.add_error(GradingPolicyIssue('grading_policy.json', msg=msg))

    # Check the GRADE_CUTOFFS entry
    if "GRADE_CUTOFFS" not in grading_policy:
        errorstore.add_error(GradingPolicyIssue('grading_policy.json', msg="'GRADE_CUTOFFS' entry not found in grading policy"))
    elif not isinstance(grading_policy["GRADE_CUTOFFS"], dict):
        errorstore.add_error(GradingPolicyIssue('grading_policy.json', msg="'GRADE_CUTOFFS' entry not in grading policy is not a dictionary"))
    else:
        # Object to entries that are not 'Pass', 'A', 'B', 'C', or 'D'
        clean = True
        for entry in grading_policy["GRADE_CUTOFFS"]:
            if entry in ['Pass', 'A', 'B', 'C', 'D']:
                # Validate the entries
                if not (isinstance(grading_policy["GRADE_CUTOFFS"][entry], int) or isinstance(grading_policy["GRADE_CUTOFFS"][entry], float)):
                    msg = f"'{entry}' entry is not a number in the GRADE_CUTOFFS part of the grading policy"
                    errorstore.add_error(GradingPolicyIssue('grading_policy.json', msg=msg))
                elif grading_policy["GRADE_CUTOFFS"][entry] > 1 or grading_policy["GRADE_CUTOFFS"][entry] < 0:
                    # We sometimes use 'Pass' < 0 to hide the pass line on the plot
                    if not (entry == 'Pass' and grading_policy["GRADE_CUTOFFS"][entry] < 0):
                        msg = f"'{entry}' entry is not between 0 and 1 in the GRADE_CUTOFFS part of the grading policy"
                        errorstore.add_error(GradingPolicyIssue('grading_policy.json', msg=msg))
            else:
                clean = False
                msg = f"'{entry}' is not allowed in the GRADE_CUTOFFS part of the grading policy"
                errorstore.add_error(GradingPolicyIssue('grading_policy.json', msg=msg))

        if clean:
            if len(grading_policy["GRADE_CUTOFFS"]) == 0:
                # Make sure it's not empty
                msg = "GRADE_CUTOFFS entry in grading policy should not be empty"
                errorstore.add_error(GradingPolicyIssue('grading_policy.json', msg=msg))
            elif 'Pass' in grading_policy["GRADE_CUTOFFS"] and len(grading_policy["GRADE_CUTOFFS"]) > 1:
                # If we have a 'Pass' entry, make sure there's no letters too
                msg = "GRADE_CUTOFFS should have either 'Pass' or letters, not both"
                errorstore.add_error(GradingPolicyIssue('grading_policy.json', msg=msg))
            elif 'Pass' not in grading_policy["GRADE_CUTOFFS"]:
                # We're doing letter grades. Make sure they're all present and in order.
                if 'D' in grading_policy["GRADE_CUTOFFS"]:
                    required = ['A', 'B', 'C', 'D']
                elif 'C' in grading_policy["GRADE_CUTOFFS"]:
                    required = ['A', 'B', 'C']
                elif 'B' in grading_policy["GRADE_CUTOFFS"]:
                    required = ['A', 'B']
                else:
                    # We only have 'A'. Should be 'Pass' instead of 'A'.
                    required = ['A']
                    msg = "GRADE_CUTOFFS should use 'Pass' instead of 'A'"
                    errorstore.add_error(GradingPolicyIssue('grading_policy.json', msg=msg))
                current = 1.1
                for entry in required:
                    if entry not in grading_policy["GRADE_CUTOFFS"]:
                        # Make sure all higher letter grades are present
                        msg = f"GRADE_CUTOFFS is missing '{entry}'"
                        errorstore.add_error(GradingPolicyIssue('grading_policy.json', msg=msg))
                    elif grading_policy["GRADE_CUTOFFS"][entry] >= current:
                        # Make sure the cutoffs are decreasing in value
                        msg = f"GRADE_CUTOFFS is entry '{entry}' is not decreasing"
                        errorstore.add_error(GradingPolicyIssue('grading_policy.json', msg=msg))
                    else:
                        current = grading_policy["GRADE_CUTOFFS"][entry]



# -*- coding: utf-8 -*-
"""
parsing_exceptions.py

Contains exception definitions for global course parsing
"""
from olxcleaner.exceptions import CourseError, ErrorLevel

class MissingURLName(CourseError):
    """A tag is missing the `url_name` attribute. edX will provide a garbage 32-character name for you, but everything is cleaner if you provide a nice name yourself."""
    _level = ErrorLevel.WARNING

    def __init__(self, filename, **kwargs):
        super().__init__(filename)
        self._description = f"The tag {kwargs['edxobj']} has no url_name."

class DuplicateURLName(CourseError):
    """Two tags have the same `url_name` attribute. This can lead to the wrong content loading, and seriously impedes this program's error analysis."""
    _level = ErrorLevel.ERROR

    def __init__(self, filename, **kwargs):
        super().__init__(filename)
        self._description = (f"Duplicate url_name found: '{kwargs['url_name']}' appears as <{kwargs['tag1']}> in "
                             f"{kwargs['file1']} and also as <{kwargs['tag2']}> in {kwargs['file2']}")

class MissingDisplayName(CourseError):
    """A tag is missing the `display_name` attribute. edX will fill a generic name for you."""
    _level = ErrorLevel.WARNING

    def __init__(self, filename, **kwargs):
        super().__init__(filename)
        self._description = f"The tag {kwargs['edxobj']} is missing the display_name attribute."

class ExtraDisplayName(CourseError):
    """A tag has a `display_name` attribute when it shouldn't."""
    _level = ErrorLevel.WARNING

    def __init__(self, filename, **kwargs):
        super().__init__(filename)
        self._description = f"The tag {kwargs['edxobj']} has an erroneous display_name attribute."

class BadPolicyFormat(CourseError):
    """The policy file didn't have the expected structure."""
    _level = ErrorLevel.ERROR

    def __init__(self, filename, **kwargs):
        super().__init__(filename)
        self._description = f"The policy file is not a dictionary of values"

class PolicyRefNotFound(CourseError):
    """The policy file references an object that doesn't exist."""
    _level = ErrorLevel.WARNING

    def __init__(self, filename, **kwargs):
        super().__init__(filename)
        self._description = f"The policy file refers to <{kwargs['objtype']} url_name='{kwargs['url_name']}'> which does not exist in the course structure"

class WrongObjectType(CourseError):
    """The policy file references an object of one type, but that object is found in the course with another type."""
    _level = ErrorLevel.ERROR

    def __init__(self, filename, **kwargs):
        super().__init__(filename)
        self._description = f"The policy file refers to a <{kwargs['objtype']}> tag with url_name '{kwargs['url_name']}'. However, that url_name points to a <{kwargs['objtypefound']}> tag."

class BadEntry(CourseError):
    """The policy file contains an entry that is not a dictionary."""
    _level = ErrorLevel.ERROR

    def __init__(self, filename, **kwargs):
        super().__init__(filename)
        self._description = f"The policy file entry for <{kwargs['objtype']} url_name='{kwargs['url_name']}'> is not a dictionary"

class SettingOverride(CourseError):
    """The policy file is overriding a setting specified in a file."""
    _level = ErrorLevel.WARNING

    def __init__(self, filename, **kwargs):
        super().__init__(filename)
        self._description = f"The policy file entry for <{kwargs['objtype']} url_name='{kwargs['url_name']}'> is overriding the setting for '{kwargs['setting']}'"

class GradingPolicyIssue(CourseError):
    """A catch-all error for issues in the grading policy."""
    _level = ErrorLevel.ERROR

    def __init__(self, filename, **kwargs):
        super().__init__(filename)
        self._description = kwargs['msg']

class InvalidSetting(CourseError):
    """A setting has been set to an invalid value."""
    _level = ErrorLevel.ERROR

    def __init__(self, filename, **kwargs):
        super().__init__(filename)
        self._description = kwargs['msg']

class DateOrdering(CourseError):
    """A date setting appears out of order with another date setting."""
    _level = ErrorLevel.WARNING

    def __init__(self, filename, **kwargs):
        super().__init__(filename)
        self._description = kwargs['msg']

class Obsolete(CourseError):
    """The way this object has been set up is obsolete."""
    _level = ErrorLevel.INFO

    def __init__(self, filename, **kwargs):
        super().__init__(filename)
        self._description = kwargs['msg']

class LTIError(CourseError):
    """There appears to be an error in the way that an LTI component is being invoked."""
    _level = ErrorLevel.ERROR

    def __init__(self, filename, **kwargs):
        super().__init__(filename)
        self._description = kwargs['msg']

class MissingFile(CourseError):
    """A file appears to be missing from the static directory."""
    _level = ErrorLevel.WARNING

    def __init__(self, filename, **kwargs):
        super().__init__(filename)
        self._description = f"The {kwargs['edxobj']} tag contains a reference to a missing static file: {kwargs['missing_file']}"

class BadJumpToLink(CourseError):
    """An internal jump_to_id link points to a url_name that doesn't exist."""
    _level = ErrorLevel.WARNING

    def __init__(self, filename, **kwargs):
        super().__init__(filename)
        self._description = f"The {kwargs['edxobj']} tag contains a link to a url_name that doesn't exist: {kwargs['link']}"

class BadCourseLink(CourseError):
    """An internal /course/ link points to a location that doesn't exist."""
    _level = ErrorLevel.WARNING

    def __init__(self, filename, **kwargs):
        super().__init__(filename)
        self._description = f"The {kwargs['edxobj']} tag contains a link to a location that doesn't exist: {kwargs['link']}"

class DuplicateID(CourseError):
    """A discussion ID is duplicated. This leads to the discussion forums randomly telling students that threads have been deleted."""
    _level = ErrorLevel.ERROR

    def __init__(self, filename, **kwargs):
        super().__init__(filename)
        self._description = f"The {kwargs['edxobj1']} tag and the {kwargs['edxobj1']} tag both use the same discussion id: {kwargs['disc_id']}"



# -*- coding: utf-8 -*-
"""
Loader submodule

Contains all methods used to load an edX course (XML and policy)
"""
from olxcleaner.loader.xml import load_course
from olxcleaner.loader.policy import load_policy



# -*- coding: utf-8 -*-
"""
errorstore.py

Structures to accumulate and track errors in a course
"""
from collections import Counter

class ErrorStore(object):
    """Class to store all errors in an import"""

    def __init__(self, ignorelist=None):
        self.errors = []
        self.ignorelist = ignorelist if ignorelist else []

    def add_error(self, error):
        """Add an error to the list (but only if not ignored)"""
        if error.name not in self.ignorelist:
            self.errors.append(error)

    def return_error(self, error_level):
        """Returns True if the highest level error is at least error_level"""
        max_level = -1
        if self.errors:
            max_level = max([error.level_val for error in self.errors])
        return max_level >= error_level

    def summary(self):
        """
        Returns a dictionary keyed by error levels containing a counter
        of errors at that level keyed by error names
        """
        # Get a count of each level of error
        levelcounter = Counter([error.level for error in self.errors])
        # For each level, get a count of each type of error
        errors = {}
        for level in levelcounter:
            errors[level] = Counter([error.name for error in self.errors if error.level == level])
        return errors



# -*- coding: utf-8 -*-
"""
validate.py

Workhorse function that validates an OLX course
"""
import os

from olxcleaner.errorstore import ErrorStore
from olxcleaner.loader import load_course, load_policy
from olxcleaner.parser.policy import (find_url_names, merge_policy,
                                      validate_grading_policy)
from olxcleaner.parser.slowvalidators import SlowValidator
from olxcleaner.parser.validators import GlobalValidator
from olxcleaner.utils import traverse

from olxcleaner.objects.common import EdxObject


def validate(filename, steps=8, ignore=None, allowed_xblocks=None):
    """
    Validate an OLX course by performing the given number of steps:

      * 1: Load the course
      * 2: Load the policy and grading policy
      * 3: Validate url_names
      * 4: Merge policy data with course, ensuring that all references are valid
      * 5: Validate the grading policy
      * 6: Have every object validate itself
      * 7: Parse the course for global errors
      * 8: Parse the course for global errors that may be time-consuming to detect

    :param filename: Location of course xml file or directory
    :param steps: Number of validation steps to take (1 = first only, 8 = all)
    :param ignore: List of errors to ignore
    :param allowed_xblocks: List of all allowed xblocks.
    :return: course object, errorstore object, url_names dictionary (or None if steps < 3)
    """
    if allowed_xblocks is not None:
        supported_xblocks = {cls.type for cls in EdxObject.get_subclasses()}
        allowed_xblocks = set(allowed_xblocks) - supported_xblocks

    # Create an error store
    if ignore is None:
        ignore = []
    errorstore = ErrorStore(ignore)

    # Validation Step #1: Load the course
    if os.path.isdir(filename):
        directory = os.path.join(filename)
        file = "course.xml"
    else:
        directory, file = os.path.split(filename)

    course = load_course(directory, file, errorstore, allowed_xblocks)
    if not course:
        return None, errorstore, None

    if steps > 1:
        # Validation Step #2: Load the policy files
        policy, grading_policy = load_policy(directory, course, errorstore)

    url_names = None
    if steps > 2:
        # Validation Step #3: Construct a dictionary of url_names
        url_names = find_url_names(course, errorstore)

    if steps > 3:
        # Validation Step #4: Merge policy data into object attributes
        merge_policy(policy, url_names, errorstore)

    if steps > 4:
        # Validation Step #5: Validate grading policy
        validate_grading_policy(grading_policy, errorstore)

    if steps > 5:
        # Validation Step #6: Have every object validate itself
        for edxobj in traverse(course):
            edxobj.validate(course, errorstore)

    if steps > 6:
        # Validation Step #7: Parse the course for global errors
        for validator in GlobalValidator.validators():
            validator(course, errorstore, url_names)

    if steps > 7:
        # Validation Step #8: Parse the course for global errors that are time-consuming to detect
        for validator in SlowValidator.validators():
            validator(course, errorstore, url_names)

    return course, errorstore, url_names



# -*- coding: utf-8 -*-
"""
openassessment.py

Object description for ORA assessments
"""
from olxcleaner.objects.common import EdxContent

class EdxORA(EdxContent):
    """edX openassessment object"""
    type = "openassessment"
    display_name = False
    can_be_pointer = False
    can_be_empty = False
    depth = 4

    def validate(self, course, errorstore):
        """
        Perform validation on this object.

        :param course: The course object, which may contain settings relevant to the validation of this object
        :param errorstore: An ErrorStore object to which errors should be reported
        :return: None
        """
        # There's a lot of structure here. We just want to validate the following date entries:
        #   <openassessment submission_start="2019-05-15T17:00:00" submission_due="2019-05-29T17:00:00">
        #     <assessments>
        #       <assessment start="2019-05-15T17:00:00" due="2019-06-05T17:00:00"/>
        #     </assessments>
        #   </openassessment>

        # Clean the start and due dates
        self.clean_date("submission_start", errorstore, required=True)
        self.clean_date("submission_due", errorstore, required=True)

        # Ensure dates fall in the correct order
        self.ensure_date_order(course.attributes.get("start"),
                               self.attributes.get("submission_start"),
                               errorstore,
                               same_ok=True,
                               error_msg="start date cannot be before course start date")
        self.ensure_date_order(self.attributes.get("submission_start"),
                               course.attributes.get("end"),
                               errorstore,
                               error_msg="start date must be before course end date")
        self.ensure_date_order(course.attributes.get("start"),
                               self.attributes.get("submission_due"),
                               errorstore,
                               error_msg="due date must be after course start date")
        self.ensure_date_order(self.attributes.get("submission_start"),
                               self.attributes.get("submission_due"),
                               errorstore,
                               error_msg="start date must be before due date")
        self.ensure_date_order(self.attributes.get("submission_due"),
                               course.attributes.get("end"),
                               errorstore,
                               same_ok=True,
                               error_msg="due date must be before course end date")

        # Find any assessments and check their dates
        for idx, assessment in enumerate(self.content.findall('./assessments/assessment')):
            # Process the dates into date objects
            startdate = assessment.get('start')
            duedate = assessment.get('due')
            startdate = self.convert2date(startdate, errorstore, f"assessment {idx + 1} start date")
            duedate = self.convert2date(duedate, errorstore, f"assessment {idx + 1} due date")

            # Now ensure the date orderings!
            # Note - these are not compared to the ORA dates, because they don't need to line up
            self.ensure_date_order(course.attributes.get("start"),
                                   startdate,
                                   errorstore,
                                   same_ok=True,
                                   error_msg=f"assessment {idx + 1} start date cannot be before course start date")
            self.ensure_date_order(startdate,
                                   course.attributes.get("end"),
                                   errorstore,
                                   error_msg=f"assessment {idx + 1} start date must be before course end date")
            self.ensure_date_order(course.attributes.get("start"),
                                   duedate,
                                   errorstore,
                                   error_msg=f"assessment {idx + 1} due date must be after course start date")
            self.ensure_date_order(startdate,
                                   duedate,
                                   errorstore,
                                   error_msg=f"assessment {idx + 1} start date must be before due date")
            self.ensure_date_order(duedate,
                                   course.attributes.get("end"),
                                   errorstore,
                                   same_ok=True,
                                   error_msg=f"assessment {idx + 1} due date must be before course end date")

            # TODO: Validate complete ORA schema



# -*- coding: utf-8 -*-
"""
xml_exceptions.py

Contains exception definitions for XML loading
"""
from olxcleaner.exceptions import CourseError, ErrorLevel

class CourseXMLDoesNotExist(CourseError):
    """The supplied `course.xml` file does not exist (or could not be opened)."""
    _level = ErrorLevel.ERROR

    def __init__(self, filename, **kwargs):
        super().__init__(filename)
        self._description = f"The file '{filename}' does not exist."

class InvalidXML(CourseError):
    """The specified XML file has a syntax error."""
    _level = ErrorLevel.ERROR

    def __init__(self, filename, **kwargs):
        super().__init__(filename)
        self._description = kwargs['error']

class InvalidHTML(CourseError):
    """The specified HTML file has a syntax error."""
    _level = ErrorLevel.ERROR

    def __init__(self, filename, **kwargs):
        super().__init__(filename)
        self._description = kwargs['error']

class CourseXMLName(CourseError):
    """The master file was not called `course.xml`."""
    _level = ErrorLevel.WARNING

    def __init__(self, filename, **kwargs):
        super().__init__(filename)
        self._description = f"The course file, {filename}, is not named course.xml"

class TagMismatch(CourseError):
    """A file purporting to contain a specific tag type (e.g., `problem` or `chapter`) instead contains a different tag."""
    _level = ErrorLevel.ERROR

    def __init__(self, filename, **kwargs):
        super().__init__(filename)
        self._description = (f"The file is of type <{kwargs['tag1']}> but "
                             f"opens with a <{kwargs['tag2']}> tag")

class EmptyTag(CourseError):
    """A tag was unexpectedly empty (e.g., a `chapter` tag had no children)."""
    _level = ErrorLevel.WARNING

    def __init__(self, filename, **kwargs):
        super().__init__(filename)
        self._description = f"The {kwargs['edxobj']} tag is unexpectedly empty"

class URLNameMismatch(CourseError):
    """A tag that had been pointed to by `url_name` from another file has a `url_name` of its own."""
    _level = ErrorLevel.WARNING

    def __init__(self, filename, **kwargs):
        super().__init__(filename)
        self._description = f"The opening <{kwargs['tag']}> tag has a mismatched url in url_name."

class InvalidPointer(CourseError):
    """This tag looks like it's trying to be a pointer but contains other attributes."""
    _level = ErrorLevel.ERROR

    def __init__(self, filename, **kwargs):
        super().__init__(filename)
        self._description = f"The {kwargs['edxobj']} tag looks like it's trying to be a pointer but contains other attributes."

class FileDoesNotExist(CourseError):
    """The file being pointed to does not exist."""
    _level = ErrorLevel.ERROR

    def __init__(self, filename, **kwargs):
        super().__init__(filename)
        self._description = f"The {kwargs['edxobj']} tag points to the file {kwargs['new_file']} that does not exist"

class SelfPointer(CourseError):
    """A tag appears to be pointing to itself."""
    _level = ErrorLevel.ERROR

    def __init__(self, filename, **kwargs):
        super().__init__(filename)
        self._description = f"The tag {kwargs['edxobj']} tag appears to be pointing to itself"

class PointerAlreadyPointedAt(CourseError):
    """A tag that has been pointed to and is pointing to another url_name"""
    _level = ErrorLevel.ERROR

    def __init__(self, filename, **kwargs):
        super().__init__(filename)
        self._description = f"The {kwargs['edxobj']} tag has been pointed to and is pointing to another url_name"

class UnexpectedTag(CourseError):
    """A tag was found in an inappropriate location (e.g., a `vertical` in a `chapter`), or the tag was not recognized."""
    _level = ErrorLevel.ERROR

    def __init__(self, filename, **kwargs):
        super().__init__(filename)
        self._description = f"A <{kwargs['tag']}> tag was unexpectedly found inside the {kwargs['edxobj']} tag"

class PossiblePointer(CourseError):
    """This tag looks like it isn't a pointer tag, but a file exists that it could be trying to point to. (This file is thus orphaned, as no other tag can point to it due to `url_name` clashes.)"""
    _level = ErrorLevel.WARNING

    def __init__(self, filename, **kwargs):
        super().__init__(filename)
        self._description = (f"The {kwargs['edxobj']} tag "
                             f"is not a pointer, but a file that it could point to exists ({kwargs['new_file']})")

class PossibleHTMLPointer(CourseError):
    """This HTML tag looks like it isn't a pointer tag, but a file exists that it could be trying to point to."""
    _level = ErrorLevel.WARNING

    def __init__(self, filename, **kwargs):
        super().__init__(filename)
        self._description = (f"The {kwargs['edxobj']} tag "
                             f"is not a pointer, but a file that it could point to exists ({kwargs['new_file']})")

class UnexpectedContent(CourseError):
    """A tag contains unexpected text content."""
    _level = ErrorLevel.ERROR

    def __init__(self, filename, **kwargs):
        super().__init__(filename)
        self._description = f"The {kwargs['edxobj']} tag should not contain any text ({kwargs['text'].strip()[:15]}...)"

class NonFlatURLName(CourseError):
    """A `url_name` pointer uses colon notation to point to a subdirectory. While partially supported, this is not recommended."""
    _level = ErrorLevel.WARNING

    def __init__(self, filename, **kwargs):
        super().__init__(filename)
        self._description = f"The {kwargs['edxobj']} tag uses obsolete colon notation in the url_name to point to a subdirectory"

class NonFlatFilename(CourseError):
    """A filename pointer for an HTML file uses colon notation to point to a subdirectory. While partially supported, this is not recommended."""
    _level = ErrorLevel.WARNING

    def __init__(self, filename, **kwargs):
        super().__init__(filename)
        self._description = f"The {kwargs['edxobj']} tag uses obsolete colon notation to point to a subdirectory for filename {kwargs['newfilename']}"

class DuplicateHTMLName(CourseError):
    """Two HTML tags point to the same HTML file (`filename` attribute). While this isn't obviously problematic, probably best not to do it."""
    _level = ErrorLevel.INFO

    def __init__(self, filename, **kwargs):
        super().__init__(filename)
        self._description = (f"Two html tags refer to the same HTML file (using the 'filename' attribute): "
                             f"{kwargs['htmlfilename']} is referenced in {filename} and {kwargs['file2']}")



# -*- coding: utf-8 -*-
"""
sequential.py

Object description for an OLX sequential tag
"""
from olxcleaner.objects.common import EdxObject, show_answer_list, randomize_list, show_correctness_list
from olxcleaner.parser.parser_exceptions import InvalidSetting

class EdxSequential(EdxObject):
    """edX sequential object"""
    type = "sequential"
    depth = 2
    display_name = True

    @property
    def allowed_children(self):
        return ['vertical']

    def validate(self, course, errorstore):
        """
        Perform validation on this object.

        :param course: The course object, which may contain settings relevant to the validation of this object
        :param errorstore: An ErrorStore object to which errors should be reported
        :return: None
        """
        self.validate_entry_from_allowed("rerandomize", randomize_list, errorstore)
        self.validate_entry_from_allowed("show_correctness", show_correctness_list, errorstore)
        self.validate_entry_from_allowed("showanswer", show_answer_list, errorstore)

        # Clean the start and due dates
        self.clean_date("start", errorstore)
        self.clean_date("due", errorstore)

        # Ensure dates fall in the correct order
        self.ensure_date_order(course.attributes.get("start"),
                               self.attributes.get("start"),
                               errorstore,
                               same_ok=True,
                               error_msg="start date cannot be before course start date")
        self.ensure_date_order(self.attributes.get("start"),
                               course.attributes.get("end"),
                               errorstore,
                               error_msg="start date must be before course end date")
        self.ensure_date_order(course.attributes.get("start"),
                               self.attributes.get("due"),
                               errorstore,
                               error_msg="due date must be after course start date")
        self.ensure_date_order(self.attributes.get("start"),
                               self.attributes.get("due"),
                               errorstore,
                               error_msg="start date must be before due date")
        self.ensure_date_order(self.attributes.get("due"),
                               course.attributes.get("end"),
                               errorstore,
                               same_ok=True,
                               error_msg="due date must be before course end date")

        # If this is a timed exam, make sure it's enabled in the policy
        if self.is_exam:
            if not course.attributes.get('enable_timed_exams'):
                msg = f"The tag {self} is a timed exam, but the course policy does not have 'enable_timed_exams=true'."
                errorstore.add_error(InvalidSetting(self.filenames[-1], msg=msg))

    @property
    def is_exam(self):
        """Helper routine that determines whether or not this is a timed exam"""
        entry = self.attributes.get('is_time_limited')
        if entry == "true":
            return True
        return False



# -*- coding: utf-8 -*-
"""
wiki.py

Object description for an OLX wiki tag
"""
from olxcleaner.objects.common import EdxContent


class Wiki(EdxContent):
    """wiki object"""
    type = "wiki"
    can_be_empty = True
    display_name = True

    def validate(self, course, errorstore):
        """
        Perform validation on this object.

        :param course: The course object, which may contain settings relevant to the validation of this object
        :param errorstore: An ErrorStore object to which errors should be reported
        :return: None
        """
        # TODO: Validate wiki
        pass



"""
test_validate.py

Tests the full validation pipeline
"""
from tests.helpers import assert_caught_all_errors, assert_error, assert_not_error
from tests.test_load_policy import handle_course1_errors
from tests.test_load_xml import handle_course2_errors, handle_nocourse_errors
from tests.test_parser import handle_course7_errors
from tests.test_validators import (handle_discussion_id_errors_in_10,
                                   handle_display_name_errors_in_10,
                                   handle_general_errors_in_10,
                                   handle_link_errors_in_10)

from olxcleaner import validate
from olxcleaner.loader.xml_exceptions import UnexpectedTag
from olxcleaner.parser.parser_exceptions import (DateOrdering, InvalidSetting,
                                                 LTIError, MissingFile,
                                                 MissingURLName, Obsolete)


def test_validate_nocourse():
    course, errorstore, url_names = validate("testcourses/nocourse.xml", 2)
    handle_nocourse_errors(errorstore)
    assert_caught_all_errors(errorstore)


def test_validate_course1():
    course, errorstore, url_names = validate("testcourses/testcourse1", 2)
    handle_course1_errors(errorstore)
    assert_caught_all_errors(errorstore)


def test_validate_course2():
    course, errorstore, url_names = validate("testcourses/testcourse2/coursefile.xml", 1)
    handle_course2_errors(errorstore)
    assert_caught_all_errors(errorstore)


def test_validate_course2_wiki_xblock_is_supported():
    """
    Test validate for course with wiki
    :return:
    """
    course, errorstore, url_names = validate("testcourses/testcourse2/coursefile.xml", 1)
    assert_not_error(errorstore, UnexpectedTag, 'course/mycourseurl.xml',
                 "A <wiki> tag was unexpectedly found inside the <course url_name='mycourseurl'> tag")


def test_validate_course7():
    course, errorstore, url_names = validate("testcourses/testcourse7/course.xml", 5)
    handle_course7_errors(errorstore)
    assert_caught_all_errors(errorstore)


def test_validate_course8():
    """This test includes individual component validation. Almost everything should pass."""
    course, errorstore, url_names = validate("testcourses/testcourse8", 6)
    assert_error(errorstore, InvalidSetting, 'course/mycourseurl.xml',
                 "The tag <course url_name='mycourseurl'> does not have the required setting 'course_image'.")
    assert_caught_all_errors(errorstore)


def test_validate_course9():
    """This test includes individual component validation. This similar to course8, but riddled with errors."""
    course, errorstore, url_names = validate("testcourses/testcourse9", 6)
    assert_error(errorstore, InvalidSetting, 'course/mycourseurl.xml',
                 "Unable to recognize graceperiod format in policy.")
    assert_error(errorstore, InvalidSetting, 'problem/problem.xml',
                 "The tag <problem url_name='problem'> has an invalid setting 'showanswer=bad'.")
    assert_error(errorstore, InvalidSetting, 'chapter/chapter.xml',
                 "The tag <chapter url_name='chapter' display_name='Hi there!'> has an invalid date setting for start: 'Feb 20, 2019, 17:00zzz'.")
    assert_error(errorstore, DateOrdering, 'sequential/sequential.xml',
                 "The tag <sequential url_name='sequential' display_name='Hi there!'> has a date out of order: start date cannot be before course start date")
    assert_error(errorstore, DateOrdering, 'vertical/vertical.xml',
                 "The tag <vertical url_name='vertical' display_name='Hi mom!'> has a date out of order: due date must be before course end date")
    assert_error(errorstore, DateOrdering, 'problem/problem.xml',
                 "The tag <problem url_name='problem'> has a date out of order: start date must be before due date")
    assert_error(errorstore, MissingURLName, 'vertical/vertical.xml',
                 "The tag <problem display_name='no url_name'> has no url_name.")
    assert_error(errorstore, InvalidSetting, 'vertical/vertical.xml',
                 "The tag <problem display_name='no url_name'> has an invalid setting 'showanswer=hah!'.")
    assert_error(errorstore, DateOrdering, 'vertical/vertical.xml',
                 "The tag <problem display_name='no url_name'> has a date out of order: due date must be before course end date")
    assert_error(errorstore, Obsolete, 'vertical/vertical.xml',
                 "The tag <discussion url_name='discussion'> should be included inline rather than through the discussion directory.")
    assert_error(errorstore, InvalidSetting, 'discussion/discussion.xml',
                 "The tag <discussion url_name='discussion'> does not have the required setting 'discussion_id'.")
    assert_error(errorstore, InvalidSetting, 'discussion/discussion.xml',
                 "The tag <discussion url_name='discussion'> does not have the required setting 'discussion_category'.")
    assert_error(errorstore, InvalidSetting, 'discussion/discussion.xml',
                 "The tag <discussion url_name='discussion'> does not have the required setting 'discussion_target'.")
    assert_error(errorstore, Obsolete, 'lti/lti.xml',
                 "The tag <lti url_name='lti'> should be converted to the newer lti_consumer Xblock.")
    assert_error(errorstore, InvalidSetting, 'lti/lti.xml',
                 "The tag <lti url_name='lti'> does not have the required setting 'launch_url'.")
    assert_error(errorstore, LTIError, 'lti/lti.xml',
                 "Course policy does not include an 'lti_passports' entry for 'nothere', required for <lti url_name='lti'>.")
    assert_error(errorstore, LTIError, 'lti/lti.xml',
                 "Course policy does not include the 'lti' advanced module, required for <lti url_name='lti'>.")
    assert_error(errorstore, InvalidSetting, 'vertical/vertical.xml',
                 "The tag <lti_consumer url_name='meep'> does not have the required setting 'launch_url'.")
    assert_error(errorstore, LTIError, 'vertical/vertical.xml',
                 "Course policy does not include an 'lti_passports' entry for 'nothere', required for <lti_consumer url_name='meep'>.")
    assert_error(errorstore, LTIError, 'vertical/vertical.xml',
                 "Course policy does not include the 'lti_consumer' advanced module, required for <lti_consumer url_name='meep'>.")
    assert_error(errorstore, MissingFile, 'course/mycourseurl.xml',
                 "The <course url_name='mycourseurl'> tag contains a reference to a missing static file: course_image_small.jpg")
    assert_error(errorstore, InvalidSetting, 'problem/problem.xml',
                 "The tag <problem url_name='problem'> has a negative problem weight.")
    assert_error(errorstore, InvalidSetting, 'course/mycourseurl.xml',
                 "The tag <course url_name='mycourseurl'> should have a positive number of attempts.")
    assert_error(errorstore, InvalidSetting, 'problem/problem.xml',
                 "The tag <problem url_name='problem'> should have a positive number of attempts.")
    assert_error(errorstore, InvalidSetting, 'vertical/vertical.xml',
                 "The tag <drag-and-drop-v2 url_name='studio_mess2' display_name='This is my title'> has an error in the data JSON: Expecting property name enclosed in double quotes: line 1 column 2 (char 1).")
    assert_error(errorstore, InvalidSetting, 'vertical/vertical.xml',
                 "The tag <drag-and-drop-v2 url_name='studio_mess3' display_name='This is my title'> data JSON is not a dictionary.")
    assert_error(errorstore, InvalidSetting, 'sequential/examseq.xml',
                 "The tag <sequential url_name='examseq' display_name='Exam'> is a timed exam, but the course policy does not have 'enable_timed_exams=true'.")
    assert_error(errorstore, DateOrdering, 'vertical/oravert.xml',
                 "The tag <openassessment url_name='paper-draft'> has a date out of order: assessment 1 due date must be before course end date")
    assert_caught_all_errors(errorstore)

    # Ensure that we got the scripts from the problem file
    assert set(url_names['problem'].scripts) == {'python', 'perl', 'javascript'}
    # Also make sure we detected the response types
    assert set(url_names['problem'].response_types) == {'customresponse', 'multiplechoiceresponse'}
    # as well as the input types
    assert set(url_names['problem'].input_types) == {'choicegroup', 'textline'}
    # We also found the solution
    assert url_names['problem'].has_solution

    # Make sure our exam sequential was detected
    assert url_names['examseq'].is_exam


def test_validate_course10():
    """This test includes all validation steps. The course is designed to test the validators and slow validators."""
    course, errorstore, url_names = validate("testcourses/testcourse10")
    handle_general_errors_in_10(errorstore)
    handle_display_name_errors_in_10(errorstore)
    handle_discussion_id_errors_in_10(errorstore)
    handle_link_errors_in_10(errorstore)
    assert_caught_all_errors(errorstore)


def test_validate_course11():
    """
    Tests validate when given a set of allowed xblocks.

    The test course uses currently unsupported xblocks which
    could fail the validate call.
    """
    allowed_xblocks = [
        "recommender",
        "edx_sga",
        "crowdsourcehinter",
        "done",
        "word_cloud"
    ]
    total_errors = len(allowed_xblocks)
    validate_kwargs = dict(filename="testcourses/testcourse11", steps=1)
    # recommender, edx_sga, crowdsourcehinter, done, word_cloud
    course, errorstore, url_names = validate(**validate_kwargs)
    assert len(errorstore.errors) == total_errors

    current_allowed_xblocks = []
    for xblock in allowed_xblocks:
        current_allowed_xblocks.append(xblock)
        # Do not throw error when an xblock is allowed.
        course, errorstore, url_names = validate(**validate_kwargs, allowed_xblocks=current_allowed_xblocks)
        assert len(errorstore.errors) == total_errors - len(current_allowed_xblocks)



# -*- coding: utf-8 -*-
"""
slowvalidators.py

Validation routines that act on the course as a whole,
but may require significant time to carry out
"""
from olxcleaner.objects import EdxDragAndDropV2, EdxVertical
from olxcleaner.parser.validators import GlobalValidator
from olxcleaner.utils import traverse, find_links, check_static_file_exists
from olxcleaner.parser.parser_exceptions import BadJumpToLink, BadCourseLink, MissingFile

class SlowValidator(GlobalValidator):
    """Abstract base class for time-consuming validators"""

class CheckLinks(SlowValidator):
    """Searches the course for broken internal links (including static links)"""

    def __call__(self, course, errorstore, url_names):
        for edxobj in traverse(course):
            if edxobj.content_store:
                # Find all of the special links in the object
                links = find_links(edxobj)
                # Make sure that each link has an endpoint!
                validate_links(course, url_names, links, edxobj, errorstore)
            elif isinstance(edxobj, EdxDragAndDropV2):
                # Look inside the data structure of dndv2 objects
                data = edxobj.parsed_data
                links = []
                if 'targetImg' in data:
                    links.append(data['targetImg'])
                if 'items' in data:
                    for entry in data['items']:
                        if 'imageURL' in entry:
                            links.append(entry['imageURL'])
                # Make sure that each link has an endpoint!
                validate_links(course, url_names, links, edxobj, errorstore)

def validate_links(course, url_names, links, edxobj, errorstore):
    """Takes in the links for an edxobj, and processes all links"""
    for link in links:
        if link.startswith('/jump_to_id/'):
            url_name = link[len('/jump_to_id/'):]
            if url_name not in url_names:
                errorstore.add_error(BadJumpToLink(edxobj.filenames[-1], edxobj=edxobj, link=link))

        elif link.startswith('/course/'):
            # Call a routine to handle this one
            follow_course_link(course, link, edxobj, errorstore)

        elif link.startswith('/static/'):
            # Call a routine to handle this one
            follow_static_link(course, link, edxobj, errorstore)

def follow_course_link(course, link, edxobj, errorstore):
    """
    Follow a course link, reporting an error if it's invalid.

    :param course: Course object
    :param link: Link to chase
    :param edxobj: Object with the link (for reporting purposes)
    :param errorstore: ErrorStore object (for reporting purposes)
    :return: None
    """
    # We need to split the link into pieces
    link_parts = link[len('/course/'):].split("/")

    if link_parts[0] == 'pdfbook':
        # If this is a textbook link, then we're not presently equipped to follow it
        # TODO: Check for links to textbooks
        return
    elif link_parts[0] != 'courseware':
        # Anything other than courseware can't be linked to either
        # e.g., Discussion forum links can never be checked...
        # TODO: Check for links to tabs (which can in principle be checked...)
        return

    # We now try to follow the path
    idx = 1
    current_obj = course
    last_term = False

    while True:
        # If we're out of link parts, we've followed the link as far as it requires
        if idx == len(link_parts) or last_term:
            return

        # If there's a ? in the link part, must be the last term
        # We ignore the ? instruction
        if "?" in link_parts[idx]:
            link_parts[idx] = link_parts[idx].split("?")[0]
            last_term = True

        # If the currently entry is empty, also good (must be done after checking for ?)
        if link_parts[idx] == "":
            return

        # If the current_obj is a vertical, then the next link part must be interpreted as an index
        if isinstance(current_obj, EdxVertical):
            try:
                link_idx = int(link_parts[idx])
            except ValueError:
                # Can't convert the entry to an index; link is bad
                errorstore.add_error(BadCourseLink(edxobj.filenames[-1], edxobj=edxobj, link=link))
                return

            if link_idx < 1 or link_idx > len(current_obj.children):
                # Link doesn't seem to point to somewhere that exists; link is bad
                errorstore.add_error(BadCourseLink(edxobj.filenames[-1], edxobj=edxobj, link=link))

            # Regardless of what happened, we're now done
            return
        else:
            # Scan the current object's children to look for the desired link
            for child in current_obj.children:
                if child.attributes.get('url_name') == link_parts[idx]:
                    # Follow it
                    current_obj = child
                    idx += 1
                    # Continue in the while loop
                    break
            else:
                # Didn't find something to follow, link is bad
                errorstore.add_error(BadCourseLink(edxobj.filenames[-1], edxobj=edxobj, link=link))
                return

def follow_static_link(course, link, edxobj, errorstore):
    """
    Check to see if a static file exists, reporting an error if it's invalid.

    :param course: Course object
    :param link: Link to chase
    :param edxobj: Object with the link (for reporting purposes)
    :param errorstore: ErrorStore object (for reporting purposes)
    :return: None
    """
    file_name = link[len('/static/'):]
    if not check_static_file_exists(course, file_name):
        errorstore.add_error(MissingFile(edxobj.filenames[-1],
                                         edxobj=edxobj,
                                         missing_file=link))



# -*- coding: utf-8 -*-
"""
edxcleaner.py

edX XML code validator, based on a very light wrapper around
the olxcleaner library. Despite the light touch, it
exposes all of the capabilities of the library.
"""
import sys
import argparse

from olxcleaner import validate
from olxcleaner.__version__ import version
from olxcleaner.reporting import construct_tree, report_errors, report_error_summary, report_statistics
from olxcleaner.loader.xml_exceptions import CourseXMLDoesNotExist


def handle_arguments():
    """Look after all command-line arguments"""
    parser = argparse.ArgumentParser(description="edX XML cleaner -- A validator for XML edX courses")

    # Required arguments
    # Location of course.xml
    parser.add_argument("-c", "--course", help="Location of course.xml (default=./course.xml)", default="course.xml")

    # Optional arguments
    # Output file for structure
    parser.add_argument("-t", "--tree", help="File to output course structure to")

    # Level to output structure to
    parser.add_argument("-l", "--level", help="Depth level to output structure to",
                        default=4, choices=[0, 1, 2, 3, 4], type=int)

    # Quiet mode
    parser.add_argument("-q", "--quiet", help="Quiet mode (no screen output)", action="store_true")

    # Error output
    parser.add_argument("-e", "--noerrors", help="Suppress error output", action="store_true")

    # Error summary
    parser.add_argument("-s", "--nosummary", help="Suppress error summary", action="store_true")

    # Error summary
    parser.add_argument("-S", "--stats", help="Output course statistics", action="store_true")

    # Failure level
    parser.add_argument("-f", "--failure", default=3, choices=[0, 1, 2, 3, 4], type=int,
                        help="Level of errors at which to declare failure: 0=DEBUG, 1=INFO, "
                             "2=WARNING, 3=ERROR (default), 4=NEVER")

    # Steps to run
    parser.add_argument("-p", "--steps", default=8, choices=[1, 2, 3, 4, 5, 6, 7, 8], type=int,
                        help="Validation steps to take: 1=load course, 2=load policies, 3=check url_names, "
                             "4=validate policy, 5=validate grading policy, 6=validate tags, "
                             "7=perform global validation, 8=perform detailed global validation (default)")

    # Ignore list
    parser.add_argument('-i', '--ignore', nargs='+', help='List of errors to ignore')

    # Parse the command line
    return parser.parse_args()


def main():
    """Entry point for command line instantiation"""
    # Read the command line arguments
    args = handle_arguments()

    if not args.quiet:
        print(f'edX XML cleaner {version} -- A validator for XML edX courses')
        print(f'Loading...')

    # Validate the course
    course, errorstore, url_names = validate(args.course, args.steps, args.ignore)
    
    # Check that the course exists
    if len(errorstore.errors) > 0 and isinstance(errorstore.errors[0], CourseXMLDoesNotExist):
        print(f"Error: {errorstore.errors[0].description}")
        sys.exit(1)

    # Output reports
    if not args.quiet:
        print(f'Loaded from {course.fullpath}')

        if not args.noerrors:
            error_report = report_errors(errorstore)
            if error_report:
                print()
                for line in error_report:
                    print(line)
        if not args.nosummary:
            print()
            for line in report_error_summary(errorstore):
                print(line)
        if args.stats:
            print()
            for line in report_statistics(course):
                print(line)

    # Output the structure to file
    if args.tree and course is not None:
        if not args.quiet:
            print()
            print(f"Writing structure to {args.tree}")
        with open(args.tree, 'w') as f:
            for line in construct_tree(course, args.level):
                f.write(line + "\n")

    # Exit with the appropriate error level
    if errorstore.return_error(args.failure):
        if not args.quiet:
            print()
            print(f"Done! Exiting with code 1")
        sys.exit(1)
    else:
        if not args.quiet:
            print()
            print(f"Done! Exiting with code 0")
        sys.exit(0)

if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
"""
video.py

Object description for an OLX video tag
"""
from olxcleaner.objects.common import EdxContent

class EdxVideo(EdxContent):
    """edX video object"""
    type = "video"
    can_be_empty = True
    display_name = True

    def validate(self, course, errorstore):
        """
        Perform validation on this object.

        :param course: The course object, which may contain settings relevant to the validation of this object
        :param errorstore: An ErrorStore object to which errors should be reported
        :return: None
        """
        # As video XML is such a mess with multiple formats, I'm not even going to try parsing it.
        # TODO: Validate video OLX
        pass



# -*- coding: utf-8 -*-
"""
edxreporter.py

This script crawls through a course and produces a LaTeX document
containing the course structure. If vertical names have a timestamp of
the form "(mm:ss)" in them, the total duration of videos for the
sequential will be calculated.

Recommended usage:
edxreporter > report.tex

If you get this error:
Character cannot be encoded into LaTeX: U+FEFF - `'
then you have some bad unicode in your display_name entries: Look through
the LaTeX file for {\bfseries ?}, which is what that character is converted
into.
"""
import sys
import argparse
import datetime
import re
from pylatexenc.latexencode import utf8tolatex

from olxcleaner import validate

def handle_arguments():
    """Look after all command-line arguments"""
    parser = argparse.ArgumentParser(description="edX XML Reporter")

    # Required arguments
    # Location of course.xml
    parser.add_argument("-c", "--course", help="Location of course.xml (default=./course.xml)", default="course.xml")

    # Optional arguments
    # Include url_names with verticals
    parser.add_argument("-u", "--url_names", help="Include url_names with verticals", action="store_true")

    # Parse the command line
    return parser.parse_args()

def sanitize(text):
    """
    This is a helper function that should be used to sanitize text for LaTeX output.
    You'll probably need to edit the LaTeX slightly afterwards, but it should look pretty good!
    """
    results = text
    # Fix quotation marks
    if results.startswith('"'):
        results = "``" + results[1:]
    if results.startswith("'"):
        results = "`" + results[1:]
    results = results.replace(" '", ' `')
    results = results.replace(' "', ' ``')
    results = results.replace('"', "''")
    # Fix unicode characters. Bad characters are replaced by a bold '?'
    results = utf8tolatex(results, substitute_bad_chars=True)
    # Fix math entries a^b
    results = re.sub(r'(.\^.)', r'$\1$', results)
    # Fix math entries a_b (note that utf8tolatex has converted _ to {\_})
    results = re.sub(r'(.){\\_}(.)', r'$\1_\2$', results)
    return results

# Here's the header of the LaTeX file we'll produce
header = r"""\documentclass{article}

\usepackage[english]{babel}
\usepackage{helvet}
\usepackage{microtype}
\usepackage{titlesec}
\usepackage{enumitem}
\usepackage[landscape,twocolumn]{geometry}
\usepackage{fullpage}

\special{papersize=8.5in,11in}
\setlength{\pdfpageheight}{\paperheight}
\setlength{\pdfpagewidth}{\paperwidth}
\setlength{\columnsep}{20pt}
\righthyphenmin2
\sloppy

\parskip = 1mm

\titleformat*{\section}{\large\bfseries\raggedright}
\setlist{topsep=0pt}

\begin{document}

\title{[coursetitlehere]}

\date{}

\maketitle

\vspace{-1.5cm}

Each page represents one chapter of material.

"""

# And here's the footer
footer = r"""
\end{document}"""


def main():
    """Entry point for command line instantiation"""
    # Read the command line arguments
    args = handle_arguments()

    # Load the course
    # We need XML structure + policy information, so go to step 4
    course, _, _ = validate(args.course, steps=4)

    # Make sure we have a course to crawl
    if not course:
        print("Cannot load course - aborting")
        sys.exit(1)

    # Crawl the course extracting the desired information
    chapters = []
    for chapter in course.children:
        name = chapter.attributes.get('display_name', '(Unnamed)')
        chapter_log = {'title': sanitize(name), 'sequentials': []}

        for seq in chapter.children:
            name = seq.attributes.get('display_name', '(Unnamed)')
            seq_log = {'title': sanitize(name), 'verticals': []}

            # Total up the duration of all videos in the sequential
            duration = datetime.timedelta()
            for vert in seq.children:
                name = vert.attributes.get('display_name', '(Unnamed)')
                if args.url_names:
                    url_name = vert.attributes.get('url_name', '(no url_name)')
                    seq_log['verticals'].append(f"({url_name}) {sanitize(name)}")
                else:
                    seq_log['verticals'].append(sanitize(name))

                for entry in vert.children:
                    if entry.type == "video":
                        # Search for a duration in the display_name of the form "(mins:seconds)"
                        name = entry.attributes.get('display_name', '')
                        length = re.search(r'\(([0-9]+):([0-9]+)\)', name)
                        if length:
                            minutes = int(length.group(1))
                            seconds = int(length.group(2))
                            duration += datetime.timedelta(minutes=minutes, seconds=seconds)

            if duration.total_seconds() > 0:
                # Report the total time
                hours, remainder = divmod(duration.total_seconds(), 3600)
                minutes, seconds = divmod(remainder, 60)
                hours = int(hours)
                minutes = int(minutes)
                seconds = int(seconds)
                if hours == 0:
                    seq_log['title'] += f" ({minutes}:{seconds:02})"
                else:
                    seq_log['title'] += f" ({hours}:{minutes:02}:{seconds:02})"

            chapter_log['sequentials'].append(seq_log)
        chapters.append(chapter_log)

    # Now do the output
    coursetitle = course.attributes.get('display_name', 'Unnamed Course')
    coursetitle += r" \\ Overview of Course Content"
    print(header.replace('[coursetitlehere]', coursetitle))

    # Iterate over the content
    needsnewpage = False
    for chapter in chapters:

        # Each chapter should be on its own page
        if needsnewpage:
            print("\n" + r"\newpage" + "\n")

        print(r"\section*{" + chapter['title'] + "}")

        for sequential in chapter['sequentials']:
            print("\n" + r"\section*{" + sequential['title'] + "}\n")
            print(r"\begin{itemize}")

            for vertical in sequential['verticals']:
                print(r"\item " + vertical)

            print(r"\end{itemize}")

        needsnewpage = True

    print(footer)



"""
test_load_policy.py

Tests for XML policy loading
"""
from tests.helpers import assert_error, assert_caught_all_errors

from olxcleaner.errorstore import ErrorStore
from olxcleaner.loader import load_policy, load_course
from olxcleaner.loader.policy_exceptions import (
    NoRunName,
    PolicyNotFound,
    BadPolicy
)

def test_no_policy():
    errorstore = ErrorStore()
    # Load course (needed before loading policy)
    course = load_course("testcourses/testcourse1", "course.xml", errorstore)
    assert_caught_all_errors(errorstore)
    # Load the (nonexistent) policy files
    policy, grading_policy = load_policy("testcourses/testcourse1", course, errorstore)
    handle_course1_errors(errorstore)
    assert_caught_all_errors(errorstore)

def handle_course1_errors(errorstore):
    assert_error(errorstore, PolicyNotFound, 'policies/mycourseurl/policy.json', "The policy file 'policies/mycourseurl/policy.json' was not found.")
    assert_error(errorstore, PolicyNotFound, 'policies/mycourseurl/grading_policy.json', "The policy file 'policies/mycourseurl/grading_policy.json' was not found.")

def test_no_url_name():
    errorstore = ErrorStore()
    # Load course (needed before loading policy)
    course = load_course("testcourses/testcourse4", "course.xml", errorstore)
    assert_caught_all_errors(errorstore)
    # Load the (nonexistent) policy files
    policy, grading_policy = load_policy("testcourses/testcourse4", course, errorstore)
    assert_error(errorstore, NoRunName, 'course.xml', "The course tag has no url_name.")
    assert_caught_all_errors(errorstore)

def test_bad_json():
    errorstore = ErrorStore()
    # Load course (needed before loading policy)
    course = load_course("testcourses/testcourse5", "course.xml", errorstore)
    assert_caught_all_errors(errorstore)
    # Load the policy files
    policy, grading_policy = load_policy("testcourses/testcourse5", course, errorstore)
    assert_error(errorstore, BadPolicy, 'policies/mycourseurl/policy.json', "The policy file 'policies/mycourseurl/policy.json' has invalid JSON: Expecting property name enclosed in double quotes: line 1 column 2 (char 1)")
    assert_error(errorstore, BadPolicy, 'policies/mycourseurl/grading_policy.json', "The policy file 'policies/mycourseurl/grading_policy.json' has invalid JSON: Expecting property name enclosed in double quotes: line 2 column 3 (char 4)")
    assert_caught_all_errors(errorstore)



# -*- coding: utf-8 -*-
"""
common.py

Contains abstract base classes to describe various edX objects
"""
from abc import ABC, ABCMeta, abstractmethod

import dateutil.parser
import pytz

from olxcleaner.exceptions import ClassDoesNotExist
from olxcleaner.parser.parser_exceptions import DateOrdering, InvalidSetting


class EdxObject(ABC):
    """Abstract base class for edX structure objects"""

    def __init__(self):
        """Initialize storage"""
        # Stores the children objects of this object
        self.children = []

        # Stores the attributes of this object
        self.attributes = {}

        # Stores the filenames associated with this object
        # If this file is loaded using a url_name pointer, will
        # contain two entries: [pointerfile, contentfile]
        # If not, will just contain one entry: [contentfile]
        self.filenames = []

    # Default settings

    # Can this object store content?
    content_store = False

    # Can this object be referenced using pointer tags? (if True, then can be located in the directory of the appropriate type)
    can_be_pointer = True
    # What attributes are needed to be a pointer tag?
    pointer_attr = {'url_name'}

    # What depth does this object typically appear at? (used for reporting)
    depth = 0

    # What is the name of the tag? (also the directory that stores this content type)
    type = None

    # Can this tag be empty?
    can_be_empty = False

    # Does this tag need a display_name attribute? Options are True, False and 'optional'
    display_name = False

    # Does this tag need a url_name attribute?
    needs_url_name = True

    # Is this element broken (and hence needs no further errors reported?)
    broken = False

    # Who is my parent?
    parent = None

    @property
    def allowed_children(self):  # pragma: no cover
        """
        What children object types can this object contain?
        Note: implemented as property to avoid forward reference problems.
        """
        return []

    def add_child(self, node):
        """Adds a child to this object"""
        self.children.append(node)
        # Mark the node's parent so that we can go up and down the chain of objects
        node.parent = self

    def add_attribs(self, attribs):
        """Adds to the attributes for this object"""
        self.attributes.update(attribs)

    def add_filename(self, value):
        """Adds a filename to the filename list for this object"""
        self.filenames.append(value)

    def __repr__(self):
        """Produce a string representation of this object"""
        name = self.attributes.get("display_name")
        url_name = self.attributes.get("url_name")
        result = f"<{self.type}"
        if url_name:
            result += f" url_name='{url_name}'"
        if name:
            result += f" display_name='{name}'"
        result += ">"
        return result

    def is_pointer(self, attribs=None):
        """
        Returns True if the attributes for this object imply that it
        is a pointer object, pointing to a new file

        Uses attribs if passed in, or self.attributes if not
        """
        if attribs is None:  # pragma: no cover
            attribs = self.attributes
        # Follows pointer convention in edX: is_pointer_tag in
        # https://github.com/openedx/edx-platform/blob/master/xmodule/x_module.py
        # Also needs to have no children and no text (not tested)
        return set(attribs.keys()) == self.pointer_attr and self.can_be_pointer

    @abstractmethod
    def validate(self, course, errorstore):  # pragma: no cover
        """
        Perform validation on this object.

        Objects should validate their contents other than those tags
        contained in allowed_children.

        :param course: The course object, which may contain settings relevant to the validation of this object
        :param errorstore: An ErrorStore object to which errors should be reported
        :return: None
        """
        pass

    # List of subclasses of EdxObject
    _subclasses = []

    @staticmethod
    def get_subclasses():
        """Get a list of all subclasses"""
        subclasses = set()
        work = [EdxObject]
        while work:
            parent = work.pop()
            for child in parent.__subclasses__():
                if child not in subclasses:
                    subclasses.add(child)
                    work.append(child)
        return list(subclasses)

    @staticmethod
    def get_object(object_type):
        """
        Returns a class instance that has type object_type

        :param object_type: type parameter for the desired class
        :return: Class with type object_type
        """
        # If it doesn't already exist, construct the list of subclasses
        if not EdxObject._subclasses:
            EdxObject._subclasses = EdxObject.get_subclasses()

        for cls in EdxObject._subclasses:
            if cls.type == object_type:
                return cls()
        raise ClassDoesNotExist(f"Cannot instantiate object of unknown type <{object_type}>")  # pragma: no cover

    def validate_entry_from_allowed(self, setting_name, allowed_list, errorstore, missing_ok=True):
        """
        Validate that the entry under the setting setting_name for this object is in the allowed list.
        If not, put the error in the errorstore.

        :param setting_name: The name of the setting
        :param allowed_list: List of allowed entries
        :param errorstore: ErrorStore object to store errors
        :param missing_ok: Whether the entry can be missing
        :return: None
        """
        # Check the entry
        entry = self.attributes.get(setting_name)
        if entry is None:
            if not missing_ok:  # pragma: no cover
                msg = f"The tag {self} does not have the required setting '{setting_name}'."
                errorstore.add_error(InvalidSetting(self.filenames[-1], msg=msg))
            return
        elif entry not in allowed_list:
            msg = f"The tag {self} has an invalid setting '{setting_name}={entry}'."
            errorstore.add_error(InvalidSetting(self.filenames[-1], msg=msg))

    def require_setting(self, setting_name, errorstore):
        """
        Validate that the setting setting_name has an entry for this object.
        If not, put the error in the errorstore.

        :param setting_name: The name of the setting
        :param errorstore: ErrorStore object to store errors
        :return: None
        """
        # Check the entry
        entry = self.attributes.get(setting_name)
        if entry is None:
            msg = f"The tag {self} does not have the required setting '{setting_name}'."
            errorstore.add_error(InvalidSetting(self.filenames[-1], msg=msg))

    def clean_date(self, setting_name, errorstore, required=False):
        """
        Clean the date entry in setting_name, storing it.
        If invalid, report the error, and nullify the entry.

        :param setting_name: The name of the setting
        :param errorstore: ErrorStore object to store errors
        :param required: Whether or not this setting is required
        :return: None
        """
        # Handle the required part
        if required:
            self.require_setting(setting_name, errorstore)

        # Handle the cleaning part
        # Done in a separate routine, so that that routine can be called when we don't want to overwrite setting
        date = self.attributes.get(setting_name)
        self.attributes[setting_name] = self.convert2date(date, errorstore, setting_name)

    def convert2date(self, date, errorstore, setting_name):
        """
        Returns a date interpretation of the given date.
        If there's an issue, store it in the errorstore, using setting_name in the error message.
        """
        if date is None:
            return None

        # Make sure it parses properly
        try:
            parsed_date = dateutil.parser.parse(date)
            if parsed_date.tzinfo is None or parsed_date.tzinfo.utcoffset(parsed_date) is None:
                # Apply a timezone to the date
                return pytz.utc.localize(parsed_date)
            else:
                # Shift to UTC
                return parsed_date.astimezone(pytz.utc)
        except (TypeError, ValueError):
            msg = f"The tag {self} has an invalid date setting for {setting_name}: '{date}'."
            errorstore.add_error(InvalidSetting(self.filenames[-1], msg=msg))
            return None

    def ensure_date_order(self, date1, date2, errorstore, error_msg, same_ok=False):
        """
        Ensures that the dates appear in order date1 < date2. If not, stores msg as an error in errorstore.
        If either date is None, no comparison is made.

        :param date1: First date to compare
        :param date2: Second date to compare
        :param errorstore: ErrorStore object to store errors
        :param error_msg: The message to report
        :param same_ok: Whether or not it's ok for the dates to be the same
        :return: None
        """
        if date1 is None or date2 is None:
            return

        # Check the ordering
        if same_ok:
            if date1 > date2:
                msg = f"The tag {self} has a date out of order: {error_msg}"
                errorstore.add_error(DateOrdering(self.filenames[-1], msg=msg))
        else:
            if date1 >= date2:
                msg = f"The tag {self} has a date out of order: {error_msg}"
                errorstore.add_error(DateOrdering(self.filenames[-1], msg=msg))

    def require_positive_attempts(self, errorstore):
        """
        Require that the object's "attempts" attribute is positive.

        :param errorstore: ErrorStore to report any errors
        :return: None
        """
        attempts = self.attributes.get("attempts")
        try:
            if attempts and int(attempts) < 1:
                msg = f"The tag {self} should have a positive number of attempts."
                errorstore.add_error(InvalidSetting(self.filenames[-1], msg=msg))
        except ValueError:
            msg = f"The tag {self} should have a positive number of attempts."
            errorstore.add_error(InvalidSetting(self.filenames[-1], msg=msg))


class EdxContent(EdxObject, metaclass=ABCMeta):
    """Abstract class for edX content objects"""

    # Can this object store content?
    content_store = True

    # What depth does this object typically appear at? (used for reporting)
    depth = 4

    # Contains the content of the tag, including the tag itself (except for HTML tags that reference an html file)
    content = None


# Collections of constants
show_answer_list = ["always", "answered", "attempted", "closed", "finished", "correct_or_past_due",
                    "past_due", "never", "after_attempts"]

randomize_list = ["always", "onreset", "never", "per_student"]

show_correctness_list = ['always', 'past_due', 'never']



# -*- coding: utf-8 -*-
"""
html.py

Object description for an OLX html tag
"""
from olxcleaner.objects.common import EdxContent

class EdxHtml(EdxContent):
    """edX html object"""
    type = "html"
    html_content = False  # Was the content set by slurping up an HTML file directly?
    # If True, content does not contain the wrapping html tag
    display_name = True

    def validate(self, course, errorstore):
        """
        Perform validation on this object.

        :param course: The course object, which may contain settings relevant to the validation of this object
        :param errorstore: An ErrorStore object to which errors should be reported
        :return: None
        """
        # Pretty much anything goes for an HTML file
        # Nothing to validate here
        pass



# -*- coding: utf-8 -*-
"""
entries submodule

Contains files for each external entry, as defined in setup.py.

This submodule is not intended for external use. Rather, it contains
light wrappers around the olxcleaner package.
"""



# -*- coding: utf-8 -*-
"""
lti_consumer.py

Object description for an OLX lti_consumer tag
"""
from olxcleaner.objects.common import EdxObject
from olxcleaner.parser.parser_exceptions import LTIError

class EdxLtiConsumer(EdxObject):
    """edX lti_consumer object"""
    can_be_pointer = False
    type = "lti_consumer"
    depth = 4
    can_be_empty = True
    display_name = True

    def validate(self, course, errorstore):
        """
        Perform validation on this object.

        :param course: The course object, which may contain settings relevant to the validation of this object
        :param errorstore: An ErrorStore object to which errors should be reported
        :return: None
        """
        # Check that required fields are present
        self.require_setting("lti_id", errorstore)
        if not self.attributes.get('hide_launch'):
            self.require_setting("launch_url", errorstore)

        # Check LTI passport exists in policy
        lti_id = self.attributes.get('lti_id')
        if lti_id:
            if not self.check_passports(course, lti_id):
                msg = f"Course policy does not include an 'lti_passports' entry for '{lti_id}', required for {self}."
                errorstore.add_error(LTIError(self.filenames[-1], msg=msg))

        # Check that lti_consumer is in the course policy as an advanced module
        if course.attributes.get('advanced_modules') is None or "lti_consumer" not in course.attributes.get('advanced_modules'):
            msg = f"Course policy does not include the 'lti_consumer' advanced module, required for {self}."
            errorstore.add_error(LTIError(self.filenames[-1], msg=msg))

    @staticmethod
    def check_passports(course, lti_id):
        """
        Checks to see if an LTI passport is in the course.
        
        :param course: The course object, which may contain settings relevant to the validation of this object
        :param lti_id: Passport to search for.
        :return: True if the passport is in the course, otherwise False.
        """
        for passport in course.attributes.get('lti_passports', []):
            if passport.startswith(lti_id + ":"):
                return True
        return False



# -*- coding: utf-8 -*-
"""
validators.py

Validation routines that act on the course as a whole
"""
import inspect
from abc import ABC, abstractmethod
from olxcleaner.objects import EdxDiscussion
from olxcleaner.utils import traverse
from olxcleaner.parser.parser_exceptions import (
    MissingDisplayName,
    ExtraDisplayName,
    DuplicateID
)

class GlobalValidator(ABC):
    """
    Abstract base class describing global validation routines.
    Only the __call__ method needs to be implemented.
    """

    @abstractmethod
    def __call__(self, course, errorstore, url_names):
        """
        Abstract method to perform validation

        :param course: EdxCourse object with a loaded course
        :param errorstore: ErrorStore object where errors are reported
        :param url_names: Dictionary of url_name to objects
        :return: None
        """

    @classmethod
    def validators(cls):
        """
        Yield an iterable of instances of immediate subclasses of this class
        """
        for child in cls.__subclasses__():
            # Only attempt instantiation if not an abstract base class
            if not inspect.isabstract(child):
                yield child()

class CheckDisplayNames(GlobalValidator):
    """Searches the course for missing display_name attributes"""

    def __call__(self, course, errorstore, url_names):
        for edxobj in traverse(course):
            display_name = edxobj.attributes.get('display_name')
            if edxobj.display_name is True and (display_name is None or display_name == ""):
                errorstore.add_error(MissingDisplayName(edxobj.filenames[-1], edxobj=edxobj))
            elif edxobj.display_name is False and display_name is not None:
                errorstore.add_error(ExtraDisplayName(edxobj.filenames[-1], edxobj=edxobj))

class CheckDiscussionIDs(GlobalValidator):
    """Searches the course for duplicate discussion_id entries in discussion blocks"""

    def __call__(self, course, errorstore, url_names):
        discussion_ids = {}
        for edxobj in traverse(course):
            if isinstance(edxobj, EdxDiscussion):
                disc_id = edxobj.attributes.get('discussion_id')
                if disc_id:
                    if disc_id in discussion_ids:
                        errorstore.add_error(DuplicateID(edxobj.filenames[-1],
                                                         edxobj1=edxobj,
                                                         edxobj2=discussion_ids[disc_id],
                                                         disc_id=disc_id))
                    else:
                        discussion_ids[disc_id] = edxobj



"""
test_parser.py

Test course parsing routines
"""
from tests.helpers import assert_error, assert_caught_all_errors

from olxcleaner.errorstore import ErrorStore
from olxcleaner.loader import load_policy, load_course
from olxcleaner.parser.policy import find_url_names, merge_policy, validate_grading_policy
from olxcleaner.parser.parser_exceptions import (
    MissingURLName,
    DuplicateURLName,
    BadPolicyFormat,
    PolicyRefNotFound,
    WrongObjectType,
    BadEntry,
    SettingOverride,
    GradingPolicyIssue
)

def test_url_names():
    errorstore = ErrorStore()

    # Load course
    course = load_course("testcourses/testcourse1", "course.xml", errorstore)
    assert_caught_all_errors(errorstore)

    # Make a dictionary of url_names
    url_names = find_url_names(course, errorstore)
    assert_caught_all_errors(errorstore)

    expected = [
        'mycourseurl',
        'chapter',
        'sequential',
        'vertical',
        'html'
    ]
    for i in expected:
        assert i in url_names

def test_course6_url():
    errorstore = ErrorStore()

    # Load course
    course = load_course("testcourses/testcourse6", "course.xml", errorstore)
    assert_caught_all_errors(errorstore)

    # Load policy
    policy, grading_policy = load_policy("testcourses/testcourse6", course, errorstore)
    assert_caught_all_errors(errorstore)

    # Make a dictionary of url_names
    url_names = find_url_names(course, errorstore)

    # Merge the policy file
    merge_policy(policy, url_names, errorstore)

    # Handle the errors
    handle_course6_errors(errorstore)
    assert_caught_all_errors(errorstore)

def handle_course6_errors(errorstore):
    assert_error(errorstore, MissingURLName, 'course/mycourseurl.xml', "The tag <vertical display_name='vertical name'> has no url_name.")
    assert_error(errorstore, MissingURLName, 'course/mycourseurl.xml', "The tag <vertical display_name='vertical name'> has no url_name.")
    assert_error(errorstore, DuplicateURLName, 'course/mycourseurl.xml', "Duplicate url_name found: 'html' appears as <html> in course/mycourseurl.xml and also as <html> in course/mycourseurl.xml")
    assert_error(errorstore, DuplicateURLName, 'course/mycourseurl.xml', "Duplicate url_name found: 'html' appears as <html> in course/mycourseurl.xml and also as <html> in course/mycourseurl.xml")
    assert_error(errorstore, DuplicateURLName, 'course/mycourseurl.xml', "Duplicate url_name found: 'html2' appears as <html> in course/mycourseurl.xml and also as <html> in course/mycourseurl.xml")
    assert_error(errorstore, BadPolicyFormat, 'policy.json', "The policy file is not a dictionary of values")

def test_course7_url():
    errorstore = ErrorStore()

    # Load course
    course = load_course("testcourses/testcourse7", "course.xml", errorstore)
    assert_caught_all_errors(errorstore)

    # Load policy
    policy, grading_policy = load_policy("testcourses/testcourse7", course, errorstore)
    assert_caught_all_errors(errorstore)

    # Make a dictionary of url_names
    url_names = find_url_names(course, errorstore)

    # Merge the policy file
    merge_policy(policy, url_names, errorstore)
    # Ensure that settings were indeed merged
    assert(url_names['sequential2'].attributes['setting'])

    # Validate the grading policy
    validate_grading_policy(grading_policy, errorstore)

    # Handle the errors
    handle_course7_errors(errorstore)
    assert_caught_all_errors(errorstore)

def handle_course7_errors(errorstore):
    assert_error(errorstore, DuplicateURLName, 'course/mycourseurl.xml', "Duplicate url_name found: 'html4' appears as <html> in course/mycourseurl.xml and also as <html> in course/mycourseurl.xml")
    assert_error(errorstore, PolicyRefNotFound, 'policy.json', "The policy file refers to <problem url_name='noexist'> which does not exist in the course structure")
    assert_error(errorstore, WrongObjectType, 'policy.json', "The policy file refers to a <video> tag with url_name 'html4'. However, that url_name points to a <html> tag.")
    assert_error(errorstore, BadEntry, 'policy.json', "The policy file entry for <vertical url_name='vertical3'> is not a dictionary")
    assert_error(errorstore, SettingOverride, 'policy.json', "The policy file entry for <sequential url_name='sequential2'> is overriding the setting for 'setting'")
    assert_error(errorstore, GradingPolicyIssue, 'grading_policy.json', "GRADE_CUTOFFS is missing 'A'")

def test_grading_policy():
    # Test grading policy entries in isolation
    errorstore = ErrorStore()

    # Set up the policy
    policy = {}
    # Validate it
    validate_grading_policy(policy, errorstore)
    # Handle errors
    assert_error(errorstore, GradingPolicyIssue, 'grading_policy.json', "'GRADER' entry not found in grading policy")
    assert_error(errorstore, GradingPolicyIssue, 'grading_policy.json', "'GRADE_CUTOFFS' entry not found in grading policy")
    assert_caught_all_errors(errorstore)

    # Set up the policy
    policy = {
        'GRADER': {},
        'GRADE_CUTOFFS': []
    }
    # Validate it
    validate_grading_policy(policy, errorstore)
    # Handle errors
    assert_error(errorstore, GradingPolicyIssue, 'grading_policy.json', "'GRADER' entry not in grading policy is not a list")
    assert_error(errorstore, GradingPolicyIssue, 'grading_policy.json', "'GRADE_CUTOFFS' entry not in grading policy is not a dictionary")
    assert_caught_all_errors(errorstore)

    # Set up the policy
    policy = {
        'GRADER': [
            {
                "drop_count": 2,
                "min_count": 12,
                "type": "Lab",
                "short_label": 15,
                "weight": 0.15
            },
            {
                "drop_count": 2,
                "min_count": 12,
                "type": "Something",
                "weight": 0.15
            },
        ],
        'GRADE_CUTOFFS': {
            'Pass': -1
        }
    }
    # Validate it
    validate_grading_policy(policy, errorstore)
    # Handle errors
    assert_error(errorstore, GradingPolicyIssue, 'grading_policy.json', "'weight' settings do not add up to 1")
    assert_error(errorstore, GradingPolicyIssue, 'grading_policy.json', "'short_label' setting is not a string for entry 1 in the grading policy")
    assert_caught_all_errors(errorstore)

    # Set up the policy
    policy = {
        'GRADER': [
            {
                "drop_count": 'hi',
                "min_count": 'hi',
                "type": 0,
                "weight": 'hi'
            },
            {
                "drop_count": -1,
                "min_count": 0,
                "type": "Something",
                "weight": 1.1
            },
            {
            },
        ],
        'GRADE_CUTOFFS': {
            'Pass': 2
        }
    }
    # Validate it
    validate_grading_policy(policy, errorstore)
    # Handle errors
    assert_error(errorstore, GradingPolicyIssue, 'grading_policy.json',
                 "'drop_count' setting is not an integer for entry 1 in the grading policy")
    assert_error(errorstore, GradingPolicyIssue, 'grading_policy.json',
                 "'min_count' setting is not an integer for entry 1 in the grading policy")
    assert_error(errorstore, GradingPolicyIssue, 'grading_policy.json',
                 "'type' setting is not a string for entry 1 in the grading policy")
    assert_error(errorstore, GradingPolicyIssue, 'grading_policy.json',
                 "'weight' setting is not a number between 0 and 1 for entry 1 in the grading policy")
    assert_error(errorstore, GradingPolicyIssue, 'grading_policy.json',
                 "'drop_count' setting is negative for entry 2 in the grading policy")
    assert_error(errorstore, GradingPolicyIssue, 'grading_policy.json',
                 "'min_count' setting is less than 1 for entry 2 in the grading policy")
    assert_error(errorstore, GradingPolicyIssue, 'grading_policy.json',
                 "'weight' setting is not a number between 0 and 1 for entry 2 in the grading policy")
    assert_error(errorstore, GradingPolicyIssue, 'grading_policy.json',
                 "'drop_count' setting is omitted for entry 3 in the grading policy")
    assert_error(errorstore, GradingPolicyIssue, 'grading_policy.json',
                 "'min_count' setting is omitted for entry 3 in the grading policy")
    assert_error(errorstore, GradingPolicyIssue, 'grading_policy.json',
                 "'type' setting is omitted for entry 3 in the grading policy")
    assert_error(errorstore, GradingPolicyIssue, 'grading_policy.json',
                 "'weight' setting is omitted for entry 3 in the grading policy")
    assert_error(errorstore, GradingPolicyIssue, 'grading_policy.json', "'weight' settings do not add up to 1")
    assert_error(errorstore, GradingPolicyIssue, 'grading_policy.json',
                 "'Pass' entry is not between 0 and 1 in the GRADE_CUTOFFS part of the grading policy")
    assert_caught_all_errors(errorstore)

    # Set up the policy
    policy = {
        'GRADER': [
            {
                "drop_count": 2,
                "min_count": 12,
                "type": "Something",
                "weight": 0.6
            },
            {
                "drop_count": 2,
                "min_count": 12,
                "type": "Something",
                "weight": 0.4
            },
        ],
        'GRADE_CUTOFFS': {
            'A': -1,
            'C': 3,
            'D': 0.5,
            'Pass': 'ok',
            'pass': 1
        }
    }
    # Validate it
    validate_grading_policy(policy, errorstore)
    # Handle errors
    assert_error(errorstore, GradingPolicyIssue, 'grading_policy.json', "'A' entry is not between 0 and 1 in the GRADE_CUTOFFS part of the grading policy")
    assert_error(errorstore, GradingPolicyIssue, 'grading_policy.json', "'C' entry is not between 0 and 1 in the GRADE_CUTOFFS part of the grading policy")
    assert_error(errorstore, GradingPolicyIssue, 'grading_policy.json', "'pass' is not allowed in the GRADE_CUTOFFS part of the grading policy")
    assert_error(errorstore, GradingPolicyIssue, 'grading_policy.json', "'Pass' entry is not a number in the GRADE_CUTOFFS part of the grading policy")
    assert_error(errorstore, GradingPolicyIssue, 'grading_policy.json', "Assessment type 'Something' appears multiple times in the grading policy")
    assert_caught_all_errors(errorstore)


    # Set up the policy
    policy = {
        'GRADER': [
            {
                "drop_count": 2,
                "min_count": 12,
                "type": "Something",
                "weight": 1
            },
        ],
        'GRADE_CUTOFFS': {
            'A': 0,
            'C': 0.2,
            'D': 0.5,
            'Pass': 0.6,
        }
    }
    # Validate it
    validate_grading_policy(policy, errorstore)
    # Handle errors
    assert_error(errorstore, GradingPolicyIssue, 'grading_policy.json', "GRADE_CUTOFFS should have either 'Pass' or letters, not both")
    assert_caught_all_errors(errorstore)


    # Set up the policy
    policy = {
        'GRADER': [
            {
                "drop_count": 2,
                "min_count": 12,
                "type": "Something",
                "weight": 1
            },
        ],
        'GRADE_CUTOFFS': {
            'A': 0,
            'C': 0.2,
            'D': 0.5
        }
    }
    # Validate it
    validate_grading_policy(policy, errorstore)
    # Handle errors
    assert_error(errorstore, GradingPolicyIssue, 'grading_policy.json', "GRADE_CUTOFFS is missing 'B'")
    assert_error(errorstore, GradingPolicyIssue, 'grading_policy.json', "GRADE_CUTOFFS is entry 'C' is not decreasing")
    assert_error(errorstore, GradingPolicyIssue, 'grading_policy.json', "GRADE_CUTOFFS is entry 'D' is not decreasing")
    assert_caught_all_errors(errorstore)


    # Set up the policy
    policy = {
        'GRADER': [
            {
                "drop_count": 2,
                "min_count": 12,
                "type": "Something",
                "weight": 1
            },
        ],
        'GRADE_CUTOFFS': {
            'A': 0,
            'B': 0.1,
            'C': 0.2,
            'D': 0.5
        }
    }
    # Validate it
    validate_grading_policy(policy, errorstore)
    # Handle errors
    assert_error(errorstore, GradingPolicyIssue, 'grading_policy.json', "GRADE_CUTOFFS is entry 'B' is not decreasing")
    assert_error(errorstore, GradingPolicyIssue, 'grading_policy.json', "GRADE_CUTOFFS is entry 'C' is not decreasing")
    assert_error(errorstore, GradingPolicyIssue, 'grading_policy.json', "GRADE_CUTOFFS is entry 'D' is not decreasing")
    assert_caught_all_errors(errorstore)

    # Set up the policy
    policy = {
        'GRADER': [
        ],
        'GRADE_CUTOFFS': {
        }
    }
    # Validate it
    validate_grading_policy(policy, errorstore)
    # Handle errors
    assert_error(errorstore, GradingPolicyIssue, 'grading_policy.json', "GRADER entry in grading policy should not be empty")
    assert_error(errorstore, GradingPolicyIssue, 'grading_policy.json', "GRADE_CUTOFFS entry in grading policy should not be empty")
    assert_caught_all_errors(errorstore)

    # Set up the policy
    policy = {
        'GRADER': [
            {
                "drop_count": 2,
                "min_count": 12,
                "type": "Something",
                "weight": 1
            },
        ],
        'GRADE_CUTOFFS': {
            'A': 1,
            'C': 0.1
        }
    }
    # Validate it
    validate_grading_policy(policy, errorstore)
    # Handle errors
    assert_error(errorstore, GradingPolicyIssue, 'grading_policy.json', "GRADE_CUTOFFS is missing 'B'")
    assert_caught_all_errors(errorstore)

    # Set up the policy
    policy = {
        'GRADER': [
            {
                "drop_count": 2,
                "min_count": 12,
                "type": "Something",
                "weight": 1
            },
        ],
        'GRADE_CUTOFFS': {
            'B': 1,
            'C': 0.1
        }
    }
    # Validate it
    validate_grading_policy(policy, errorstore)
    # Handle errors
    assert_error(errorstore, GradingPolicyIssue, 'grading_policy.json', "GRADE_CUTOFFS is missing 'A'")
    assert_caught_all_errors(errorstore)


    # Set up the policy
    policy = {
        'GRADER': [
            {
                "drop_count": 2,
                "min_count": 12,
                "type": "Something",
                "weight": 1
            },
        ],
        'GRADE_CUTOFFS': {
            'B': 1
        }
    }
    # Validate it
    validate_grading_policy(policy, errorstore)
    # Handle errors
    assert_error(errorstore, GradingPolicyIssue, 'grading_policy.json', "GRADE_CUTOFFS is missing 'A'")
    assert_caught_all_errors(errorstore)


    # Set up the policy
    policy = {
        'GRADER': [
            {
                "drop_count": 2,
                "min_count": 12,
                "type": "Something",
                "weight": 1
            },
        ],
        'GRADE_CUTOFFS': {
            'A': 1
        }
    }
    # Validate it
    validate_grading_policy(policy, errorstore)
    # Handle errors
    assert_error(errorstore, GradingPolicyIssue, 'grading_policy.json', "GRADE_CUTOFFS should use 'Pass' instead of 'A'")
    assert_caught_all_errors(errorstore)



# -*- coding: utf-8 -*-
"""
objects subpackage

Contains classes to describe various edX objects

Note that each edX object has its own file

Use the EdxCourse.get_object method to create objects of appropriate tags
"""
from olxcleaner.objects.common import EdxObject
from olxcleaner.objects.chapter import EdxChapter
from olxcleaner.objects.course import EdxCourse
from olxcleaner.objects.discussion import EdxDiscussion
from olxcleaner.objects.html import EdxHtml
from olxcleaner.objects.lti import EdxLti
from olxcleaner.objects.lti_consumer import EdxLtiConsumer
from olxcleaner.objects.problem import EdxProblem
from olxcleaner.objects.sequential import EdxSequential
from olxcleaner.objects.vertical import EdxVertical
from olxcleaner.objects.video import EdxVideo
from olxcleaner.objects.draganddropv2 import EdxDragAndDropV2
from olxcleaner.objects.openassessment import EdxORA
from olxcleaner.objects.wiki import Wiki



# -*- coding: utf-8 -*-
"""
reporting.py

Contains methods used to report on an analyzed course
"""
from collections import Counter
from olxcleaner.utils import traverse
from olxcleaner.objects import EdxProblem, EdxSequential
from olxcleaner.exceptions import ErrorLevel

def report_statistics(course):
    """Report course statistics, formatted in a list."""
    result = []

    # If things are borked, then get out
    if course is None or course.broken:  # pragma: no cover
        return result

    # Compute statistics
    typecounter, exams, response_types, input_types, python_problems, problem_solutions = compute_statistics(course)

    # Now print to screen
    result.append("Number of each type of object:")
    for entry in typecounter:
        result.append(f"  - {entry}: {typecounter[entry]}")

    result.append(f"Number of exams: {exams}")

    if typecounter['problem'] > 0:
        result.append("Problem statistics:")
        result.append(f"    Number of problems: {typecounter['problem']}")
        result.append(f"    Number of problems with solutions: {problem_solutions}")
        result.append(f"    Number of problems with python scripts: {python_problems}")
        result.append(f"    response_types used*:")
        for entry in response_types:
            result.append(f"      - {entry}: {response_types[entry]}")
        result.append(f"    input_types used*:")
        for entry in input_types:
            result.append(f"      - {entry}: {input_types[entry]}")
        result.append(f"    * Multiple uses within a single problem only count once")

    return result

def compute_statistics(course):
    """
    Compute course statistics, returned in a number of variables.

    :param course: A validated course.
    :return: typecounter (dict),
             exams (int),
             response_types (dict),
             input_types (dict),
             python_problems (int),
             problem_solutions (int)
    """
    # Track how many of each type of object we have
    typecounter = Counter()

    # Track the number of sequentials that are time-limited (i.e., exams)
    exams = 0

    # Track the number of each type of response type and input type in problems
    response_types = Counter()
    input_types = Counter()

    # Track the number of problems using python scripts
    python_problems = 0

    # Track the number of problems with solutions
    problem_solutions = 0

    for edxobj in traverse(course):
        # Track the object type
        typecounter[edxobj.type] += 1

        if isinstance(edxobj, EdxSequential):
            # Check for exams
            if edxobj.is_exam:
                exams += 1

        elif isinstance(edxobj, EdxProblem):
            # Check for the 4 problem settings
            if edxobj.has_solution:
                problem_solutions += 1

            for response_type in edxobj.response_types:
                response_types[response_type] += 1

            for input_type in edxobj.input_types:
                input_types[input_type] += 1

            if 'python' in edxobj.scripts:
                python_problems += 1

    return typecounter, exams, response_types, input_types, python_problems, problem_solutions

def report_errors(errorstore):
    """Gives a simple report of all errors that were found, returned as a list"""
    result = []
    if errorstore.errors:
        for error in errorstore.errors:
            result.append((error.filename, f"{error.level} {error.name} ({error.filename}): {error.description}"))
    result.sort()  # Uses the filename for ordering
    wanted = [report for (filename, report) in result]
    return wanted

def report_error_summary(errorstore):
    """Reports summary statistics on the errors found, returned as a list"""
    result = ["Summary:"]
    counter = errorstore.summary()
    for level in ErrorLevel:
        if level.name in counter:
            result.append(f"{level.name}s: {sum(counter[level.name].values())}")
            for name, num in sorted(counter[level.name].items()):
                result.append(f"    {name}: {num}")
    if not counter:  # pragma: no cover
        result.append("No errors found!")
    return result

def construct_tree(course, maxdepth=None):
    """
    Constructs a tree version of the course structure, formatted as a list.

    :param course: EdxCourse object
    :param maxdepth: Maximum depth to display:
        0 = course
        1 = chapter
        2 = sequential
        3 = vertical
        4 = content
    :return: list containing tree structure
    """
    result = []
    _construct_tree(course, result, maxdepth)
    return result

def _construct_tree(obj, output, maxdepth, indent=0):
    """
    Recursively writes a tree version of the course structure to a list.

    :param obj: Current object to display
    :param output: List to append output to
    :param maxdepth: The maximum depth to display
    :param indent: Current indentation level (used in recursion)
    :return: None
    """
    if maxdepth is not None and obj.depth > maxdepth:
        return
    output.append(f'{" " * indent * 4}{obj}')
    for child in obj.children:
        _construct_tree(child, output, maxdepth, indent + 1)



# -*- coding: utf-8 -*-
"""
policy_exceptions.py

Contains exception definitions for policy file loading
"""
from olxcleaner.exceptions import CourseError, ErrorLevel

class NoRunName(CourseError):
    """The course tag has no `url_name`, and hence no run name. This is a required parameter for a course."""
    _level = ErrorLevel.ERROR
    
    def __init__(self, filename, **kwargs):
        super().__init__(filename)
        self._description = f"The course tag has no url_name."

class PolicyNotFound(CourseError):
    """A policy file was not found."""
    _level = ErrorLevel.ERROR

    def __init__(self, filename, **kwargs):
        super().__init__(filename)
        self._description = f"The policy file '{filename}' was not found."

class BadPolicy(CourseError):
    """A policy file was not valid JSON."""
    _level = ErrorLevel.ERROR

    def __init__(self, filename, **kwargs):
        super().__init__(filename)
        self._description = f"The policy file '{filename}' has invalid JSON: {kwargs['msg']}"



# -*- coding: utf-8 -*-
"""
discussion.py

Object description for an OLX discussion tag
"""
from olxcleaner.objects.common import EdxObject
from olxcleaner.parser.parser_exceptions import Obsolete

class EdxDiscussion(EdxObject):
    """edX discussion object"""
    type = "discussion"
    depth = 4
    can_be_empty = True
    needs_url_name = False
    display_name = 'optional'

    def validate(self, course, errorstore):
        """
        Perform validation on this object.

        :param course: The course object, which may contain settings relevant to the validation of this object
        :param errorstore: An ErrorStore object to which errors should be reported
        :return: None
        """
        # Check for obsolete invocation
        if len(self.filenames) == 2:
            msg = f"The tag {self} should be included inline rather than through the discussion directory."
            errorstore.add_error(Obsolete(self.filenames[0], msg=msg))

        # Check for required attributes
        self.require_setting("discussion_id", errorstore)
        self.require_setting("discussion_category", errorstore)
        self.require_setting("discussion_target", errorstore)



"""
test_validators.py

Tests for validators and slow validators
"""
from olxcleaner import validate
from olxcleaner.loader.xml_exceptions import TagMismatch
from olxcleaner.parser.validators import CheckDisplayNames, CheckDiscussionIDs
from olxcleaner.parser.slowvalidators import CheckLinks
from tests.helpers import assert_caught_all_errors, assert_error
from olxcleaner.parser.parser_exceptions import (MissingDisplayName, ExtraDisplayName, Obsolete,
                                                 DuplicateID, BadJumpToLink, MissingFile, BadCourseLink)
from olxcleaner.utils import find_links

def test_display_names():
    # Perform all steps on course 10 up to validation steps
    course, errorstore, url_names = validate("testcourses/testcourse10", 6)
    handle_general_errors_in_10(errorstore)
    assert_caught_all_errors(errorstore)

    # Check the display names
    validator = CheckDisplayNames()
    validator(course, errorstore, url_names)

    # Catch everything!
    handle_display_name_errors_in_10(errorstore)
    assert_caught_all_errors(errorstore)

def handle_general_errors_in_10(errorstore):
    assert_error(errorstore, TagMismatch, 'chapter/broken_chapter.xml', "The file is of type <chapter> but opens with a <sequential> tag")
    assert_error(errorstore, Obsolete, 'vertical/dndvert.xml', "The tag <lti url_name='lti2' display_name='something'> should be converted to the newer lti_consumer Xblock.")
    assert_error(errorstore, TagMismatch, 'html/tagmismatch.xml', "The file is of type <html> but opens with a <div> tag")


def handle_display_name_errors_in_10(errorstore):
    assert_error(errorstore, MissingDisplayName, 'course/mycourseurl.xml', "The tag <course url_name='mycourseurl'> is missing the display_name attribute.")
    assert_error(errorstore, MissingDisplayName, 'course/mycourseurl.xml', "The tag <html url_name='html'> is missing the display_name attribute.")
    assert_error(errorstore, ExtraDisplayName, 'vertical/oravert.xml', "The tag <openassessment url_name='paper-draft' display_name='Hah!'> has an erroneous display_name attribute.")

def test_discussion_ids():
    # Perform all steps on course 10 up to validation steps
    course, errorstore, url_names = validate("testcourses/testcourse10", 6)
    handle_general_errors_in_10(errorstore)
    assert_caught_all_errors(errorstore)

    # Check the discussion IDs
    validator = CheckDiscussionIDs()
    validator(course, errorstore, url_names)

    # Catch everything!
    handle_discussion_id_errors_in_10(errorstore)
    assert_caught_all_errors(errorstore)

def handle_discussion_id_errors_in_10(errorstore):
    assert_error(errorstore, DuplicateID, 'course/mycourseurl.xml', "The <discussion display_name='Here's a discussion'> tag and the <discussion display_name='Here's a discussion'> tag both use the same discussion id: Me")

def test_link_checking():
    """Checks for broken internal links"""
    # Perform all steps on course 10 up to validation steps
    course, errorstore, url_names = validate("testcourses/testcourse10", 6)
    handle_general_errors_in_10(errorstore)
    assert_caught_all_errors(errorstore)

    # Check the link-finding routines
    links = find_links(url_names['linktest'])
    assert set(links) == {'/course/courseware/testing',
                          '/course/courseware/broken_chapter',
                          '/course/courseware/chapter/sequential',
                          '/course/courseware/chapter/sequential/',
                          '/course/courseware/chapter/sequential/vertical',
                          '/course/courseware/chapter/sequential/vertical/',
                          '/course/courseware/chapter/sequential/vertical/1',
                          '/course/courseware/chapter/sequential/vertical/10',
                          '/course/courseware/chapter/sequential/vertical/0',
                          '/course/courseware/chapter/sequential/vertical/html',
                          '/course/courseware/chapter/sequential/vertical/1?last_child',
                          '/course/courseware/chapter/sequential/vertical/?last_child',
                          '/course/courseware/chapter/sequential/vertical?last_child',
                          '/static/testing.png',
                          '/static/testing2.css',
                          '/jump_to_id/oravert',
                          '/jump_to_id/oravert2',
                          '/static/image.png',
                          '/course/discussion/somewhere',
                          '/course/pdfbook/0/chapter/9/11'}

    # Check the links
    validator = CheckLinks()
    validator(course, errorstore, url_names)

    # Catch everything!
    handle_link_errors_in_10(errorstore)
    assert_caught_all_errors(errorstore)

def handle_link_errors_in_10(errorstore):
    assert_error(errorstore, BadJumpToLink, 'html/linktest.xml', "The <html url_name='linktest' display_name='Testing links'> tag contains a link to a url_name that doesn't exist: /jump_to_id/oravert2")
    assert_error(errorstore, MissingFile, 'html/linktest.xml', "The <html url_name='linktest' display_name='Testing links'> tag contains a reference to a missing static file: /static/testing2.css")
    assert_error(errorstore, MissingFile, 'html/linktest.xml', "The <html url_name='linktest' display_name='Testing links'> tag contains a reference to a missing static file: /static/testing.png")
    assert_error(errorstore, BadCourseLink, 'html/linktest.xml', "The <html url_name='linktest' display_name='Testing links'> tag contains a link to a location that doesn't exist: /course/courseware/testing")
    assert_error(errorstore, BadCourseLink, 'html/linktest.xml', "The <html url_name='linktest' display_name='Testing links'> tag contains a link to a location that doesn't exist: /course/courseware/chapter/sequential/vertical/10")
    assert_error(errorstore, BadCourseLink, 'html/linktest.xml', "The <html url_name='linktest' display_name='Testing links'> tag contains a link to a location that doesn't exist: /course/courseware/chapter/sequential/vertical/0")
    assert_error(errorstore, BadCourseLink, 'html/linktest.xml', "The <html url_name='linktest' display_name='Testing links'> tag contains a link to a location that doesn't exist: /course/courseware/chapter/sequential/vertical/html")
    assert_error(errorstore, MissingFile, 'vertical/dndvert.xml', "The <problem url_name='dndtest' display_name='Mwa'> tag contains a reference to a missing static file: /static/ex34_dnd_sol.png")
    assert_error(errorstore, MissingFile, 'vertical/dndvert.xml', "The <problem url_name='dndtest' display_name='Mwa'> tag contains a reference to a missing static file: /static/ex34_dnd.png")
    assert_error(errorstore, MissingFile, 'vertical/dndvert.xml', "The <problem url_name='dndtest' display_name='Mwa'> tag contains a reference to a missing static file: /static/ex34_dnd_label1.png")
    assert_error(errorstore, MissingFile, 'vertical/dnd2vert.xml', "The <drag-and-drop-v2 url_name='studio_mess' display_name='This is my title'> tag contains a reference to a missing static file: /static/ex34_dnd.png")
    assert_error(errorstore, MissingFile, 'vertical/dnd2vert.xml', "The <drag-and-drop-v2 url_name='studio_mess' display_name='This is my title'> tag contains a reference to a missing static file: /static/ex34_dnd_label1.png")



# -*- coding: utf-8 -*-
"""
policyloader.py

Routines to load the policy files of an edX course
"""
import os
from os.path import isfile
import json
from olxcleaner.loader.policy_exceptions import (
    NoRunName,
    PolicyNotFound,
    BadPolicy
)

default_grading_policy = {
    "GRADER": [
        {
            "drop_count": 0,
            "min_count": 1,
            "short_label": "HW",
            "type": "Homework",
            "weight": 1
        },
    ],
    "GRADE_CUTOFFS": {
        "Pass": 0.5
    }
}

def load_policy(directory, course, errorstore):
    """
    Loads the policy file for a course. If loading fails, empty
    dictionaries are returned.

    :param directory: Path for course.xml (or equivalent)
    :param course: EdxCourse object from parsed course
    :param errorstore: ErrorStore object to store errors
    :return: policy, grading_policy objects
    """
    # Get the run name for the course
    runname = course.attributes.get("url_name")
    if runname is None:
        errorstore.add_error(NoRunName(course.filenames[0]))
        return {}, {}

    # Construct filenames for policy files
    policyfile = os.path.join("policies", runname, "policy.json")
    gradingfile = os.path.join("policies", runname, "grading_policy.json")

    # Load the policy files
    policy = load_json(directory, policyfile, errorstore)
    grading_policy = load_json(directory, gradingfile, errorstore)
    if not grading_policy:
        grading_policy = default_grading_policy

    # Return the results
    return policy, grading_policy

def load_json(directory, filename, errorstore):
    """
    Load json from a file, storing any loading errors in the errorstore

    :param directory: Course directory
    :param filename: File to load
    :param errorstore: ErrorStore object to store errors
    :return: Contents of json file
    """
    fullfile = os.path.join(directory, filename)

    if not isfile(fullfile):
        errorstore.add_error(PolicyNotFound(filename))
        return {}

    try:
        with open(fullfile) as f:
            return json.load(f)
    except json.decoder.JSONDecodeError as err:
        errorstore.add_error(BadPolicy(filename, msg=str(err)))
        return {}



# -*- coding: utf-8 -*-
"""
utils.py

Contains utility functions used in various parts of the library
"""
import os
from os.path import isfile
import re

def traverse(edxobj):
    """
    Returns a generator that traverses a given object and all its children recursively

    This allows routines to traverse an entire course by using:
    for obj in traverse(course):
        (do something with every obj in the course)

    :param edxobj: EdxObject to traverse
    :return: Generator of EdxObjects
    """
    # Don't traverse broken objects
    if edxobj.broken:
        return

    # Generate this object
    yield edxobj

    # Generate every object generated by the children of this object
    for child in edxobj.children:
        for entry in traverse(child):
            yield entry

def check_static_file_exists(course, filename):
    """
    Checks that a given file exists in the static directory.

    :param course: Course object, needed to extract directory
    :param filename: Filename to look for
    :return: True/False
    """
    fullpath = os.path.join(course.directory, "static", filename)
    return isfile(fullpath)


# Copied from the edx-platform xmodule.fields library
TIMEDELTA_REGEX = re.compile(r'^((?P<days>\d+?) day(?:s?))?(\s)?((?P<hours>\d+?) hour(?:s?))?(\s)?((?P<minutes>\d+?) minute(?:s)?)?(\s)?((?P<seconds>\d+?) second(?:s)?)?$')

def validate_graceperiod(entry):
    """Returns True if entry is a valid grace period"""
    if entry is None:
        return True
    parts = TIMEDELTA_REGEX.match(entry)
    if not parts:
        return False
    return True

def find_links(edxobj):
    """Find all internal links in the given object"""
    links = []

    # This is the list of all attributes we will scan for internal links
    url_attributes = ['link', 'src', 'href', 'img', 'icon', 'preprocessorSrc']
    # This is the list of special links we will look for
    internal_links = ['/static/', '/course/', '/jump_to_id/']

    # Search for all elements that have the desired attributes
    if edxobj.content is not None:  # Empty objects are stored as None
        for attrib in url_attributes:
            for tag in edxobj.content.findall(f".//*[@{attrib}]"):
                link = tag.get(attrib)
                if link:
                    for special in internal_links:
                        if link.startswith(special):
                            links.append(link)

    return links



# -*- coding: utf-8 -*-
"""
setup.py

Contains setup information for the olxcleaner library.
"""
import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="olxcleaner",
    version="0.3.0",
    author="Jolyon Bloomfield",
    author_email="jolyon@mit.edu",
    description="Tool to scan Open edX courses for various errors",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/openedx/olxcleaner",
    license='LICENSE',
    packages=setuptools.find_packages(exclude=['tests']),
    classifiers=[
        "Intended Audience :: Education",
        "Topic :: Education :: Computer Aided Instruction (CAI)",
        "Development Status :: 4 - Beta",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    keywords='edx',
    install_requires=[
       'lxml',
       'python-dateutil',
       'pytz',
       'pylatexenc'
    ],
    python_requires='~=3.6',
    entry_points={
        'console_scripts': [
            'edx-cleaner=olxcleaner.entries.edxcleaner:main',
            'edx-reporter=olxcleaner.entries.edxreporter:main'
        ],
    }
)



# -*- coding: utf-8 -*-
"""
draganddropv2.py

Object description for the "new" drag and drop interface
"""
import json
from olxcleaner.objects.common import EdxObject
from olxcleaner.parser.parser_exceptions import InvalidSetting

class EdxDragAndDropV2(EdxObject):
    """edX drag-and-drop-v2 object"""
    type = "drag-and-drop-v2"
    display_name = 'optional'
    can_be_pointer = False
    can_be_empty = True
    depth = 4

    def __init__(self):
        # Do standard initialization
        super().__init__()
        # Add in extra objects
        self.parsed_data = {}

    def validate(self, course, errorstore):
        """
        Perform validation on this object.

        :param course: The course object, which may contain settings relevant to the validation of this object
        :param errorstore: An ErrorStore object to which errors should be reported
        :return: None
        """
        # The data for this Xblock is exported as json in the data field
        datafields = self.attributes.get('data')
        if datafields is None:
            return

        # Make sure that the field is valid, and save it for future validation
        try:
            parsed_data = json.loads(datafields)
            self.parsed_data = parsed_data
            if not isinstance(self.parsed_data, dict):
                msg = f"The tag {self} data JSON is not a dictionary."
                errorstore.add_error(InvalidSetting(self.filenames[-1], msg=msg))
        except json.decoder.JSONDecodeError as err:
            msg = f"The tag {self} has an error in the data JSON: {err}."
            errorstore.add_error(InvalidSetting(self.filenames[-1], msg=msg))
            return



# -*- coding: utf-8 -*-
"""
problem.py

Object description for an OLX problem tag
"""
from olxcleaner.objects.common import EdxContent, show_answer_list, randomize_list, show_correctness_list
from olxcleaner.parser.parser_exceptions import InvalidSetting

response_types = ['coderesponse', 'numericalresponse', 'formularesponse', 'customresponse', 'schematicresponse',
                  'externalresponse', 'imageresponse', 'optionresponse', 'symbolicresponse', 'stringresponse',
                  'choiceresponse', 'multiplechoiceresponse', 'truefalseresponse', 'annotationresponse',
                  'choicetextresponse']

input_types = ['optioninput', 'choicegroup', 'radiogroup', 'checkboxgroup', 'jsinput', 'textline', 'filesubmission',
               'codeinput', 'textbox', 'matlabinput', 'schematic', 'imageinput', 'crystallography', 'vsepr_input',
               'chemicalequationinput', 'formulaequationinput', 'drag_and_drop_input', 'editamoleculeinput',
               'designprotein2dinput', 'editageneinput', 'annotationinput', 'radiotextgroup', 'checkboxtextgroup']

class EdxProblem(EdxContent):
    """edX problem object"""
    type = "problem"
    display_name = True

    def __init__(self):
        # Do standard initialization
        super().__init__()
        # Add in extra objects
        self.scripts = []
        self.response_types = []
        self.input_types = []
        self.has_solution = False

    def validate(self, course, errorstore):
        """
        Perform validation on this object.

        :param course: The course object, which may contain settings relevant to the validation of this object
        :param errorstore: An ErrorStore object to which errors should be reported
        :return: None
        """
        self.validate_entry_from_allowed("rerandomize", randomize_list, errorstore)
        self.validate_entry_from_allowed("show_correctness", show_correctness_list, errorstore)
        self.validate_entry_from_allowed("showanswer", show_answer_list, errorstore)

        # Clean the start and due dates
        self.clean_date("start", errorstore)
        self.clean_date("due", errorstore)

        # Ensure dates fall in the correct order
        self.ensure_date_order(course.attributes.get("start"),
                               self.attributes.get("start"),
                               errorstore,
                               same_ok=True,
                               error_msg="start date cannot be before course start date")
        self.ensure_date_order(self.attributes.get("start"),
                               course.attributes.get("end"),
                               errorstore,
                               error_msg="start date must be before course end date")
        self.ensure_date_order(course.attributes.get("start"),
                               self.attributes.get("due"),
                               errorstore,
                               error_msg="due date must be after course start date")
        self.ensure_date_order(self.attributes.get("start"),
                               self.attributes.get("due"),
                               errorstore,
                               error_msg="start date must be before due date")
        self.ensure_date_order(self.attributes.get("due"),
                               course.attributes.get("end"),
                               errorstore,
                               same_ok=True,
                               error_msg="due date must be before course end date")

        # Ensure that problem weight isn't negative
        weight = self.attributes.get("weight")
        if weight and float(weight) < 0:
            msg = f"The tag {self} has a negative problem weight."
            errorstore.add_error(InvalidSetting(self.filenames[-1], msg=msg))

        # Ensure that number of attempts is None or positive
        self.require_positive_attempts(errorstore)

        # Detect scripts
        self.scripts = self.detect_scripts()

        # Detect response types
        self.response_types = self.detect_response_types()

        # Detect input types
        self.input_types = self.detect_input_types()

        # Detect solution
        self.has_solution = self.detect_solution()

    def detect_solution(self):
        """
        Find any solutions present in the problem

        :return: True/False
        """
        if self.content is None:
            return False

        # Does there exist at least one <solution> tag?
        if self.content.find('.//solution') is not None:
            return True
        return False

    def detect_response_types(self):
        """
        Locate and identify all response types used in the problem

        :return: List of response types used
        """
        tags = []
        if self.content is None:
            return []

        for rtype in response_types:
            # Does there exist at least one of these tags?
            if self.content.find('.//' + rtype) is not None:
                tags.append(rtype)
        return tags

    def detect_input_types(self):
        """
        Locate and identify all input types used in the problem

        :return: List of input types used
        """
        tags = []
        if self.content is None:
            return []

        for itype in input_types:
            # Does there exist at least one of these tags?
            if self.content.find('.//' + itype) is not None:
                tags.append(itype)
        return tags

    def detect_scripts(self):
        """
        Locate and identify the language of all scripts used in the problem

        :return: List of script languages used
        """
        # This code is modified from the edx-platform repository
        codetypes = set()
        if self.content is None:
            return []

        for script in self.content.findall('.//script'):
            stype = script.get('type')
            # Code is contained in script.text
            if stype:
                if 'javascript' in stype:
                    codetypes.add('javascript')
                    continue
                elif 'perl' in stype:
                    codetypes.add('perl')
                    continue
            # If not javascript or perl, we assume python (even if not present)
            codetypes.add('python')

        return list(codetypes)



"""
test_load_xml.py

Tests for XML course loading
"""
from lxml import etree
from tests.helpers import assert_error, assert_caught_all_errors

from olxcleaner.loader.xml import load_course
from olxcleaner.errorstore import ErrorStore
from olxcleaner.loader.xml_exceptions import (
    CourseXMLDoesNotExist,
    InvalidXML,
    CourseXMLName,
    TagMismatch,
    SelfPointer,
    FileDoesNotExist,
    NonFlatURLName,
    NonFlatFilename,
    InvalidPointer,
    UnexpectedTag,
    InvalidHTML,
    URLNameMismatch,
    UnexpectedContent,
    EmptyTag,
    PossiblePointer,
    PossibleHTMLPointer,
    DuplicateHTMLName,
    PointerAlreadyPointedAt)

def test_no_course():
    errorstore = ErrorStore()
    course = load_course("testcourses", "nocourse.xml", errorstore)
    handle_nocourse_errors(errorstore)
    assert_caught_all_errors(errorstore)

def handle_nocourse_errors(errorstore):
    assert_error(errorstore, CourseXMLDoesNotExist, 'testcourses/nocourse.xml', 'The file \'testcourses/nocourse.xml\' does not exist.')

def test_course1():
    """Make sure that things load properly when everything is ok"""
    errorstore = ErrorStore()
    course = load_course("testcourses/testcourse1", "course.xml", errorstore)
    assert errorstore.errors == []

    assert course.attributes == {'url_name': 'mycourseurl', 'org': 'myorg', 'course': 'mycourse'}

    [chapter] = course.children
    assert chapter.attributes == {'url_name': 'chapter', 'display_name': 'chapter name'}

    [sequential] = chapter.children
    assert sequential.attributes == {'url_name': 'sequential', 'display_name': 'display name'}

    [vertical] = sequential.children
    assert vertical.attributes == {'url_name': 'vertical', 'display_name': 'vertical name'}

    [html] = vertical.children
    assert html.attributes == {'url_name': 'html', 'display_name': 'html name'}
    assert '<p>Test course</p>' in etree.tostring(html.content, pretty_print=True).decode()

def test_course2():
    """Test for XML loading errors"""
    errorstore = ErrorStore()
    course = load_course("testcourses/testcourse2", "coursefile.xml", errorstore)
    handle_course2_errors(errorstore)
    assert_caught_all_errors(errorstore)

def handle_course2_errors(errorstore):
    assert_error(errorstore, CourseXMLName, 'coursefile.xml', 'The course file, coursefile.xml, is not named course.xml')
    assert_error(errorstore, URLNameMismatch, 'sequential/mysequential.xml', "The opening <sequential> tag has a mismatched url in url_name.")
    assert_error(errorstore, FileDoesNotExist, 'sequential/mysequential.xml', "The <vertical url_name='myverticalnone'> tag points to the file vertical/myverticalnone.xml that does not exist")
    assert_error(errorstore, EmptyTag, 'vertical/myvertical4.xml', "The <vertical url_name='myvertical4' display_name='Hello'> tag is unexpectedly empty")
    assert_error(errorstore, EmptyTag, 'sequential/mysequential.xml', "The <vertical> tag is unexpectedly empty")
    assert_error(errorstore, UnexpectedContent, 'vertical/myvertical4-2.xml', "The <vertical url_name='myvertical4-2' display_name='Hello'> tag should not contain any text (but not really...)")
    assert_error(errorstore, InvalidXML, 'vertical/myvertical5.xml', 'Opening and ending tag mismatch: vertical line 1 and chapter, line 2, column 11')
    assert_error(errorstore, UnexpectedTag, 'vertical/myvertical6.xml', "A <vertical> tag was unexpectedly found inside the <vertical url_name='myvertical6' display_name='Hello'> tag")
    assert_error(errorstore, UnexpectedTag, 'sequential/mysequential.xml', "A <chapter> tag was unexpectedly found inside the <vertical> tag")
    assert_error(errorstore, TagMismatch, 'vertical/myvertical7.xml', 'The file is of type <vertical> but opens with a <chapter> tag')
    assert_error(errorstore, SelfPointer, 'vertical/myvertical8.xml', "The tag <vertical url_name='myvertical8'> tag appears to be pointing to itself")
    assert_error(errorstore, UnexpectedContent, 'vertical/myvertical10.xml', "The <vertical url_name='myvertical10' display_name='something'> tag should not contain any text (Here is some co...)")
    assert_error(errorstore, UnexpectedContent, 'sequential/mysequential.xml', "The <vertical> tag should not contain any text (Here's some bad...)")
    assert_error(errorstore, InvalidHTML, 'html/html3.html', 'Unexpected end tag : b, line 1, column 34')
    assert_error(errorstore, PossiblePointer, 'vertical/myvertical1.xml', "The <problem url_name='problem2'> tag is not a pointer, but a file that it could point to exists (problem/problem2.xml)")
    assert_error(errorstore, PossibleHTMLPointer, 'html/html6.xml', "The <html url_name='html6'> tag is not a pointer, but a file that it could point to exists (html/html3.html)")
    assert_error(errorstore, PossibleHTMLPointer, 'vertical/myvertical3.xml', "The <html> tag is not a pointer, but a file that it could point to exists (html/html3.html)")
    assert_error(errorstore, NonFlatURLName, 'vertical/myvertical2.xml', "The <video url_name='stuff:video2'> tag uses obsolete colon notation in the url_name to point to a subdirectory")
    assert_error(errorstore, NonFlatFilename, 'html/html2.xml', "The <html url_name='html2'> tag uses obsolete colon notation to point to a subdirectory for filename html/stuff/html2.html")
    assert_error(errorstore, NonFlatFilename, 'vertical/myvertical3.xml', "The <html> tag uses obsolete colon notation to point to a subdirectory for filename html/stuff/here.html")
    assert_error(errorstore, FileDoesNotExist, 'html/html4.xml', "The <html url_name='html4'> tag points to the file html/htmlnotexist.html that does not exist")
    assert_error(errorstore, FileDoesNotExist, 'vertical/myvertical3.xml', "The <html> tag points to the file html/nonexistant.html that does not exist")
    assert_error(errorstore, DuplicateHTMLName, 'vertical/myvertical3.xml', "Two html tags refer to the same HTML file (using the 'filename' attribute): html/html7.html is referenced in vertical/myvertical3.xml and html/html7.xml")
    assert_error(errorstore, EmptyTag,'problem/problem3.xml' , "The <problem url_name='problem3' display_name='Blank Common Problem'> tag is unexpectedly empty")
    assert_error(errorstore, URLNameMismatch, 'video/video3.xml', "The opening <video> tag has a mismatched url in url_name.")
    assert_error(errorstore, URLNameMismatch, 'discussion/discussion1.xml', "The opening <discussion> tag has a mismatched url in url_name.")
    assert_error(errorstore, InvalidPointer, 'sequential/mysequential.xml', "The <vertical url_name='myvertical9' display_name='Hi there'> tag looks like it's trying to be a pointer but contains other attributes.")
    assert_error(errorstore, PointerAlreadyPointedAt, 'sequential/sequence2.xml', "The <sequential url_name='some-non-existant-sequential'> tag has been pointed to and is pointing to another url_name")
    assert_error(errorstore, FileDoesNotExist, 'chapter/mychapter.xml', "The <sequential url_name='some-other-sequential-not-found'> tag points to the file sequential/some-other-sequential-not-found.xml that does not exist")

def test_course3():
    """Test for XML error in course.xml"""
    errorstore = ErrorStore()
    course = load_course("testcourses/testcourse3", "course.xml", errorstore)
    assert_error(errorstore, InvalidXML, 'course.xml', 'attributes construct error, line 1, column 61')
    assert_caught_all_errors(errorstore)



# -*- coding: utf-8 -*-
"""
course.py

Object description for an OLX course tag
"""
from olxcleaner.objects.common import EdxObject, show_answer_list, randomize_list, show_correctness_list
from olxcleaner.parser.parser_exceptions import MissingFile, InvalidSetting
from olxcleaner.utils import check_static_file_exists, validate_graceperiod

class EdxCourse(EdxObject):
    """edX course object"""
    type = "course"
    depth = 0
    display_name = True

    # course pointer tags need three attributes:
    pointer_attr = {'url_name', 'course', 'org'}

    @property
    def allowed_children(self):
        return ["chapter", "wiki"]

    directory = None
    fullpath = None

    def savedir(self, directory, fullpath):
        """Saves the course directory and full path for future use"""
        self.directory = directory
        self.fullpath = fullpath

    def validate(self, course, errorstore):
        """
        Perform validation on this object.

        :param course: The course object, which may contain settings relevant to the validation of this object
        :param errorstore: An ErrorStore object to which errors should be reported
        :return: None
        """
        # Require some settings that need no further validation
        self.require_setting("org", errorstore)
        self.require_setting("course", errorstore)

        # Validate settings from the allowed list
        self.validate_entry_from_allowed("rerandomize", randomize_list, errorstore)
        self.validate_entry_from_allowed("show_correctness", show_correctness_list, errorstore)
        self.validate_entry_from_allowed("showanswer", show_answer_list, errorstore)

        # Handle start/end dates
        self.clean_date("start", errorstore, required=True)
        self.clean_date("end", errorstore, required=True)
        self.ensure_date_order(self.attributes.get("start"),
                               self.attributes.get("end"),
                               errorstore,
                               error_msg="start date must be before end date")

        # Handle enrollment_start/enrollment_end dates
        self.clean_date("enrollment_start", errorstore)
        self.clean_date("enrollment_end", errorstore)

        # Make sure there's a course image
        self.require_setting("course_image", errorstore)
        if self.attributes.get("course_image"):
            if not check_static_file_exists(self, self.attributes.get("course_image")):
                errorstore.add_error(MissingFile(self.filenames[-1],
                                                 edxobj=self,
                                                 missing_file=self.attributes.get("course_image")))

        # Check that the grace period is valid
        if not validate_graceperiod(self.attributes.get("graceperiod")):
            errorstore.add_error(InvalidSetting(self.filenames[-1],
                                                msg="Unable to recognize graceperiod format in policy."))

        # Ensure that the default number of attempts is None or positive
        self.require_positive_attempts(errorstore)



# -*- coding: utf-8 -*-
"""
xml.py

Routines to load the XML of an edX course into a structure
"""
import os
from os.path import isfile

from lxml import etree
from lxml.etree import XMLSyntaxError

from olxcleaner.exceptions import ClassDoesNotExist
from olxcleaner.loader.xml_exceptions import (CourseXMLDoesNotExist,
                                              CourseXMLName, DuplicateHTMLName,
                                              EmptyTag, URLNameMismatch,
                                              FileDoesNotExist, InvalidHTML,
                                              InvalidPointer, InvalidXML,
                                              NonFlatFilename, NonFlatURLName,
                                              PossibleHTMLPointer,
                                              PossiblePointer, SelfPointer,
                                              TagMismatch, UnexpectedContent,
                                              UnexpectedTag, PointerAlreadyPointedAt)
from olxcleaner.objects import EdxObject


def load_course(directory, filename, errorstore, allowed_xblocks=None):
    """
    Loads a course, given a filename for the appropriate course.xml file.

    :param directory: Path for course.xml (or equivalent)
    :param filename: Filename for course.xml (or equivalent)
    :param errorstore: ErrorStore object to store errors
    :param allowed_xblocks: List of all allowed xblocks.
    :return: EdxCourse object, or None on failure
    """
    # Ensure the file exists
    fullpath = os.path.join(directory, filename)
    if not isfile(fullpath):
        errorstore.add_error(CourseXMLDoesNotExist(fullpath))
        return

    # Check file name
    if filename != "course.xml":
        errorstore.add_error(CourseXMLName(filename))

    # Obtain the XML for the course.xml file
    try:
        tree = etree.parse(fullpath)
    except XMLSyntaxError as e:
        errorstore.add_error(InvalidXML(filename, error=e.args[0]))
        return

    # Initialize the course object
    course = EdxObject.get_object('course')

    # Load the course!
    read_course(course, tree.getroot(), directory, filename, errorstore, {}, allowed_xblocks=allowed_xblocks)

    # Save the course directory and full path in the course object
    course.savedir(directory, fullpath)

    return course


def read_course(edxobj, node, directory, filename, errorstore, htmlfiles, pointer=False, allowed_xblocks=None):
    """
    Takes in the current EdxObject, the current lxml element, and the
    current filename. Reads from the element into the object, creating
    any children for that object, and recursing into them.

    :param edxobj: The current EdxObject
    :param node: The current lxml element
    :param directory: The course directory
    :param filename: The current filename
    :param errorstore: An ErrorStore object that is collecting errors
    :param htmlfiles: A dictionary of XML filenames (value) that reference a given HTML filename (key)
    :param pointer: True if we've arrived at this node due to a pointer tag
    :param allowed_xblocks: List of all allowed xblocks.
    :return: None
    """
    # Make sure that the node matches the edxobj type
    if edxobj.type != node.tag:
        errorstore.add_error(TagMismatch(filename,
                                         tag1=edxobj.type,
                                         tag2=node.tag))
        edxobj.broken = True
        return

    # Start by copying the attributes from the node into the object
    edxobj.add_attribs(node.attrib)
    # Set the filename for the object
    edxobj.add_filename(filename)

    # Is the tag an empty tag? (contains only attributes and comments)
    nodelen = len(node)
    for child in node:
        if child.tag is etree.Comment:
            # We need to remove comment nodes from the length count
            nodelen -= 1
    empty = False
    if nodelen == 0:
        # Check for text
        if node.text is None or node.text.strip() == '':
            empty = True
            # No text, but make sure that the comment children also have no text!
            for child in node:
                if child.tail and child.tail.strip():
                    empty = False

    # Check for a pointer tag
    node_filename = os.path.splitext(os.path.split(filename)[-1])[0]
    if empty and edxobj.is_pointer(node.attrib):
        if pointer:
            node_pointing_to_itself = node.attrib.get('url_name') == node_filename
            if node_pointing_to_itself:
                # The target of a pointer tag cannot be a pointer itself
                pointer_error = SelfPointer
            else:
                # The target of a pointer has been pointed to and is unexpectedly empty.
                pointer_error = PointerAlreadyPointedAt

            errorstore.add_error(pointer_error(filename, edxobj=edxobj))
            edxobj.broken = True
            return

        # We have a valid pointer tag
        url_name = edxobj.attributes['url_name']
        if ":" in url_name:
            errorstore.add_error(NonFlatURLName(filename, edxobj=edxobj))
            url_name = url_name.replace(":", "/")
        new_file = edxobj.type + "/" + url_name + ".xml"

        # Ensure the file exists
        if not isfile(os.path.join(directory, new_file)):
            errorstore.add_error(FileDoesNotExist(filename,
                                                  edxobj=edxobj,
                                                  new_file=new_file))
            edxobj.broken = True
            return

        try:
            new_node = etree.parse(os.path.join(directory, new_file)).getroot()
        except XMLSyntaxError as e:
            errorstore.add_error(InvalidXML(new_file, error=e.args[0]))
            edxobj.broken = True
            return
        else:
            read_course(edxobj, new_node, directory, new_file, errorstore, htmlfiles, pointer=True,
                        allowed_xblocks=allowed_xblocks)
            return

    # Special case: HTML files can point to an actual HTML file with their 'filename' attribute
    if node.tag == "html" and "filename" in edxobj.attributes:
        new_filename = edxobj.attributes['filename']
        if ":" in new_filename:
            new_filename = new_filename.replace(":", "/")
            new_file = "html/" + new_filename + ".html"
            errorstore.add_error(NonFlatFilename(filename, edxobj=edxobj, newfilename=new_file))
        else:
            new_file = "html/" + new_filename + ".html"

        # If not empty, then it could be a PossibleHTMLPointer error
        if not empty:
            if isfile(os.path.join(directory, new_file)):
                errorstore.add_error(PossibleHTMLPointer(filename,
                                                         edxobj=edxobj,
                                                         new_file=new_file))
        else:
            # We are empty, so this is a good pointer
            # Ensure the file exists
            if not isfile(os.path.join(directory, new_file)):
                errorstore.add_error(FileDoesNotExist(filename,
                                                      edxobj=edxobj,
                                                      new_file=new_file))
                return

            try:
                with open(os.path.join(directory, new_file)) as f:
                    html = f.read()
                parser = etree.HTMLParser(recover=False)
                content = etree.fromstring(html, parser)
            except Exception as e:
                errorstore.add_error(InvalidHTML(new_file, error=e.args[0]))
                edxobj.broken = True
                return
            else:
                if new_filename in htmlfiles:
                    errorstore.add_error(DuplicateHTMLName(filename,
                                                           file2=htmlfiles[new_filename],
                                                           htmlfilename=new_file))
                else:
                    htmlfiles[new_filename] = filename
                edxobj.content = content
                edxobj.html_content = True
                return

    # Next, check if the tag shouldn't be empty, and hence should be a pointer
    # but for some reason was an invalid pointer
    if empty and not edxobj.can_be_empty and "url_name" in node.attrib:
        # Likely to be an invalid pointer tag due to too many attributes
        errorstore.add_error(InvalidPointer(filename, edxobj=edxobj))
        edxobj.broken = True
        return

    # At this stage, we've checked for pointer tags and associated errors

    # The pointer url should match with it's filename
    if pointer and "url_name" in node.attrib and not node.attrib.get('url_name') == node_filename:
        errorstore.add_error(URLNameMismatch(filename, tag=node.tag))

    # Is the tag unexpectedly empty?
    if empty and not edxobj.can_be_empty:
        errorstore.add_error(EmptyTag(filename, edxobj=edxobj))
        return

    # If we get here, we have a non-empty tag

    # Check to see if there is a pointer target file that is not being used
    if not pointer and 'url_name' in edxobj.attributes and edxobj.can_be_pointer:
        new_file = edxobj.type + "/" + edxobj.attributes['url_name'] + ".xml"
        if isfile(os.path.join(directory, new_file)):
            errorstore.add_error(PossiblePointer(filename,
                                                 edxobj=edxobj,
                                                 new_file=new_file))

    if edxobj.content_store:
        # Store content from content tags
        edxobj.content = node  # Can convert to text with etree.tostring(node, pretty_print=True)
    else:
        # Check for content in non-content tag
        if node.text and node.text.strip():
            errorstore.add_error(UnexpectedContent(filename,
                                                   edxobj=edxobj,
                                                   text=node.text))
        else:
            for child in node:
                if child.tail and child.tail.strip():
                    errorstore.add_error(UnexpectedContent(filename,
                                                           edxobj=edxobj,
                                                           text=child.tail))
                    break

        # Recurse into each child
        for child in node:
            child_tag = child.tag
            tag_is_allowed = check_tag_is_allowed(child_tag, edxobj.allowed_children, allowed_xblocks)
            tag_is_comment = child_tag is etree.Comment  # Ignore comments
            if tag_is_comment:
                continue
            elif tag_is_allowed:
                try:
                    newobj = EdxObject.get_object(child_tag)
                    edxobj.add_child(newobj)
                    read_course(newobj, child, directory, filename, errorstore, htmlfiles,
                                allowed_xblocks=allowed_xblocks)
                except ClassDoesNotExist:
                    # Do not do anything if an xblock is allowed but currently
                    # not supported by the library
                    pass
            else:
                errorstore.add_error(
                    UnexpectedTag(filename, tag=child_tag, edxobj=edxobj)
                )


def check_tag_is_allowed(tag, allowed_children, allowed_xblocks):
    """
    Checks if a tag is valid or not.

    :param tag: Tag name of the lxml element
    :param allowed_children: List of children/xblocks allowed by the olxcleaner
    :param allowed_xblocks: List of Xblocks that are allowed but may not be supported by the olxcleaner for validation
    """
    if allowed_xblocks is None:
        return tag in allowed_children
    return tag in allowed_xblocks or tag in allowed_children
